from __future__ import absolute_import, print_function
from qtpy import QtCore, QtWidgets, QtGui
from ScopeFoundry.logged_quantity import LQCollection#, LoggedQuantity
from collections import OrderedDict
import pyqtgraph as pg
import warnings
from ScopeFoundry.helper_funcs import get_logger_from_class, QLock
from .base_app import BaseMicroscopeApp
import time
import threading

class HardwareComponent(QtCore.QObject):
    """
    :class:`HardwareComponent`
    
    Base class for ScopeFoundry Hardware objects
    
    to subclass, implement :meth:`setup`, :meth:`connect` and :meth:`disconnect`
    
    """
    connection_succeeded = QtCore.Signal()
    connection_failed = QtCore.Signal()
    
    

    def add_logged_quantity(self, name, **kwargs):
        #lq = LoggedQuantity(name=name, **kwargs)
        #self.logged_quantities[name] = lq
        #return lq
        return self.settings.New(name, **kwargs)

    def add_operation(self, name, op_func):
        """
        Create an operation for the HardwareComponent.
        
        *op_func* is a function that will be called upon operation activation
        
        operations are typically exposed in the default ScopeFoundry gui via a pushButton
        
        :type name: str
        :type op_func: QtCore.Slot
        """
        
        self.operations[name] = op_func   
            
    def __init__(self, app: BaseMicroscopeApp, debug=False, name:str=None):
        """
        create new HardwareComponent attached to *app*
        """
        QtCore.QObject.__init__(self)

        if not hasattr(self, 'name'):
            self.name = self.__class__.__name__

        if name is not None:
            self.name = name
            
                    
        self.log = get_logger_from_class(self)

        # threading lock
        #self.lock = threading.Lock()
        #self.lock = DummyLock()
        self.lock = QLock(mode=1) # mode 0 is non-reentrant lock

        self.app = app

        #self.logged_quantities = OrderedDict()
        self.settings = LQCollection()
        self.operations = OrderedDict()

        self.connected = self.settings.New("connected", dtype=bool, 
                                           colors=['none','rgba( 0, 255, 0, 120)'],
                                           description=f'to <i>{self.name}</i> hardware if checked.')
        self.connected.updated_value[bool].connect(self.enable_connection)

        self.connect_success = False

        
        self.debug_mode = self.settings.New("debug_mode", dtype=bool, initial=debug, 
                                            colors=['none','yellow'])
        
        self.auto_thread_lock = True
        
        self.setup()

        if self.auto_thread_lock:        
            self.thread_lock_all_lq()

        self.has_been_connected_once = False
        
        self.is_connected = False
        
        self.connection_failed.connect(self.on_connection_failed)
        self.connection_succeeded.connect(self.on_connection_succeeded)
        
        self.add_operation('Reload_Code', self.reload_code)
        
    def setup(self):
        """
        Runs during __init__, before the hardware connection is established
        Should generate desired LoggedQuantities, operations
        """
        raise NotImplementedError()

    def new_control_widgets(self):
        
        self.controls_groupBox = QtWidgets.QGroupBox(self.name)
        self.controls_formLayout = QtWidgets.QFormLayout()
        self.controls_groupBox.setLayout(self.controls_formLayout)
                
        #self.connect_hardware_checkBox = QtWidgets.QCheckBox("Connect to Hardware")
        #self.controls_formLayout.addRow("Connect", self.connect_hardware_checkBox)
        #self.connect_hardware_checkBox.stateChanged.connect(self.enable_connection)

        
        self.control_widgets = OrderedDict()
        for lqname, lq in self.settings.as_dict().items():
            #: :type lq: LoggedQuantity
            if lq.choices is not None:
                widget = QtWidgets.QComboBox()
            elif lq.dtype in [int, float]:
                if lq.si:
                    widget = pg.SpinBox()
                else:
                    widget = QtWidgets.QDoubleSpinBox()
            elif lq.dtype in [bool]:
                widget = QtWidgets.QCheckBox()  
            elif lq.dtype in [str]:
                widget = QtWidgets.QLineEdit()
            lq.connect_bidir_to_widget(widget)

            # Add to formlayout
            self.controls_formLayout.addRow(lqname, widget)
            self.control_widgets[lqname] = widget
        
        self.op_buttons = OrderedDict()
        for op_name, op_func in self.operations.items(): 
            op_button = QtWidgets.QPushButton(op_name)
            op_button.clicked.connect(lambda checked, f=op_func: f())
            self.controls_formLayout.addRow(op_name, op_button)
        
        self.read_from_hardware_button = QtWidgets.QPushButton("Read From Hardware")
        self.read_from_hardware_button.clicked.connect(self.read_from_hardware)
        self.controls_formLayout.addRow("Logged Quantities:", self.read_from_hardware_button)
        
        return self.controls_groupBox
    
    def add_widgets_to_tree(self, tree):
        #tree = self.app.ui.hardware_treeWidget
        #tree = QTreeWidget()
        tree.setColumnCount(2)
        tree.setHeaderLabels(["Hardware", "Value"])

        self.tree_item = QtWidgets.QTreeWidgetItem(tree, [self.name, "o"])
        tree.insertTopLevelItem(0, self.tree_item)
        self.tree_item.setFirstColumnSpanned(False)
        self.tree_item.setForeground(1, QtGui.QColor('red'))
        
        # Add logged quantities to tree
        self.settings.add_widgets_to_subtree(self.tree_item)        
        
        # Add oepration buttons to tree        
        self.op_buttons = OrderedDict()
        for op_name, op_func in self.operations.items(): 
            op_button = QtWidgets.QPushButton(op_name)
            op_button.clicked.connect(lambda checked, f=op_func: f())
            self.op_buttons[op_name] = op_button
            #self.controls_formLayout.addRow(op_name, op_button)
            op_tree_item = QtWidgets.QTreeWidgetItem(self.tree_item, [op_name, ""])
            tree.setItemWidget(op_tree_item, 1, op_button)

        self.tree_read_from_hardware_button = QtWidgets.QPushButton("Read From\nHardware")
        self.tree_read_from_hardware_button.clicked.connect(self.read_from_hardware)
        #self.controls_formLayout.addRow("Logged Quantities:", self.read_from_hardware_button)
        self.read_from_hardware_button_tree_item = QtWidgets.QTreeWidgetItem(self.tree_item, ["Logged Quantities:", ""])
        self.tree_item.addChild(self.read_from_hardware_button_tree_item)
        tree.setItemWidget(self.read_from_hardware_button_tree_item, 1, self.tree_read_from_hardware_button)

    @QtCore.Slot()    
    def read_from_hardware(self):
        """
        Read all settings (:class:`LoggedQuantity`) connected to hardware states
        """
        for name, lq in self.settings.as_dict().items():
            if lq.has_hardware_read():
                if self.debug_mode.val: self.log.debug("read_from_hardware {}".format(name) )
                lq.read_from_hardware()
        
    
    def connect(self):
        """
        Opens a connection to hardware
        and connects :class:`LoggedQuantity` settings to related hardware 
        functions and parameters 
        """
        raise NotImplementedError()
        
        
    def disconnect(self):
        """
        Disconnects the hardware and severs hardware--:class:`LoggedQuantity` links
        """
        
        raise NotImplementedError()
    
    @QtCore.Slot(bool)
    def enable_connection(self, enable=True):
        if enable:
            try:
                self.connect()
                # start thread if needed
                if hasattr(self, 'run'):
                    self.update_thread_interrupted = False
                    self._update_thread = threading.Thread(target=self.run)
                    self._update_thread.start()

                self.connection_succeeded.emit()
            except Exception as err:
                self.connection_failed.emit()
                raise err
        else:
            print("disabling connection")
            try:
                try:
                    if hasattr(self, 'run') and hasattr(self, '_update_thread'):
                        self.update_thread_interrupted = True
                        self._update_thread.join(timeout=5.0)
                        del self._update_thread
                finally:
                    self.disconnect()
                    self.tree_item.setText(1,'X')
                    self.tree_item.setForeground(1, QtGui.QColor('red'))          
            except Exception as err:
                # disconnect failed
                self.tree_item.setText(1,'?')
                self.tree_item.setForeground(1, QtGui.QColor('red'))        
                raise err
            
    def run(self):
        if hasattr(self, 'threaded_update'):
            while not self.update_thread_interrupted:
                try:
                    self.threaded_update()
                except Exception as err:
                    print("threaded update failed", err)
                    time.sleep(1.0)
            
    def on_connection_succeeded(self):
        print(self.name, "connection succeeded!")
        self.tree_item.setText(1,'O')
        self.tree_item.setForeground(1, QtGui.QColor('green'))

            
    def on_connection_failed(self):
        print(self.name, "connection failed!")        
        self.settings.connected.update_value(False)
        self.tree_item.setText(1,'!')
        self.tree_item.setForeground(1, QtGui.QColor('red'))   
          

    @property
    def gui(self):
        warnings.warn("Hardware.gui is deprecated, use Hardware.app", DeprecationWarning)
        return self.app
    
    def web_ui(self):
        return "Hardware {}".format(self.name)
    
    
    def thread_lock_lq(self, lq):
        lq.old_lock = lq.lock
        lq.lock = self.lock
        
    def thread_lock_all_lq(self):
        for lq in self.settings.as_list():
            lq.old_lock = lq.lock
            lq.lock = self.lock
            
    def reload_code(self):
        import inspect
        import xreload
        mod = inspect.getmodule(self)
        x = xreload.xreload(mod)
        print("Reloading from code", mod, x)
        
    def New_UI(self):
        scroll_area = self.settings.New_UI(style='scroll_form')
        for n,func in self.operations.items():
            btn = QtWidgets.QPushButton(n)
            btn.clicked.connect(func)
            scroll_area.widget().layout().addRow(btn)
        read_from_hardware_button = QtWidgets.QPushButton("Read From Hardware")
        read_from_hardware_button.clicked.connect(self.read_from_hardware)
        scroll_area.widget().layout().addRow(read_from_hardware_button)
        return scroll_area


from ScopeFoundry import LQCollection, BaseApp
from qtpy import QtGui, QtWidgets

class LQWidgetTestApp(BaseApp):
    
    name = 'LQWidgetTestApp'

    def __init__(self,argv):
        BaseApp.__init__(self,argv)
        
        self.settings.New('long_string_test',  dtype=str, initial="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aliquam accumsan, ligula a tristique luctus, felis est blandit libero, nec placerat justo diam at lacus. Fusce volutpat vitae lacus non lobortis. Fusce porttitor varius placerat. Curabitur et varius urna, sit amet gravida leo. Etiam eleifend luctus erat, vel maximus libero lacinia at. Pellentesque mattis pulvinar sem, sit amet porttitor mi maximus in. Sed venenatis orci sit amet nulla luctus, vitae pulvinar neque facilisis. Donec in felis sodales libero fringilla aliquam eu non urna. Praesent ac elit ac lorem cursus aliquam eu venenatis mauris. Proin sed aliquet nunc. Duis venenatis mi dapibus.", ro=False)


        self.ui = self.settings.New_UI()
        
        long_string_test_plainTextEdit = QtWidgets.QPlainTextEdit()
        self.ui.layout().addRow("long_string_test_plainTextEdit", long_string_test_plainTextEdit)
        
        self.settings.long_string_test.connect_to_widget(long_string_test_plainTextEdit)
        
        self.console_widget.show()
        self.ui.show()

if __name__ == '__main__':
    import sys
    app = LQWidgetTestApp(sys.argv)
    app.exec_()


from qtpy.QtWidgets import (QCheckBox, QComboBox, QGridLayout, QGroupBox,
                            QLabel, QLineEdit, QPushButton, QTextEdit)


class NewHWView(QGroupBox):

    def __init__(self) -> None:
        super().__init__()
        self.setup()

    def setup(self) -> None:
        layout = QGridLayout()
        self.setLayout(layout)
        col = 0

        self.company_label = QLabel('company')
        self.company_le = QLineEdit('thorlabs')
        layout.addWidget(QLabel('company'), col, 0)
        layout.addWidget(self.company_le, col, 1, 1, 2)
        col += 1

        self.model_label = QLabel('series or model')
        self.model_le = QLineEdit('XX312_Piezo_Controller')
        layout.addWidget(self.model_label, col, 0, 1, 2)
        layout.addWidget(self.model_le, col, 1, 1, 2)
        col += 1

        self.authors_label = QLabel('authors')
        self.authors_le = QLineEdit("Benedikt Ursprung")
        layout.addWidget(self.authors_label, col, 0)
        layout.addWidget(self.authors_le, col, 1, 1, 2)
        col += 1

        self.pretty_name_label = QLabel('pretty name')
        self.pretty_name_le = QLineEdit()
        self.guess_pretty_name_btn = QPushButton('U')
        self.guess_pretty_name_btn.setMaximumWidth(15)
        layout.addWidget(self.pretty_name_label, col, 0)
        layout.addWidget(self.pretty_name_le, col, 1)
        layout.addWidget(self.guess_pretty_name_btn, col, 2)
        col += 1

        self.com_types_label = QLabel('communication type')
        self.com_types_combob = QComboBox()

        layout.addWidget(self.com_types_label, col, 0)
        layout.addWidget(self.com_types_combob, col, 1, 1, 2)
        col += 1

        self.overwirte_label = QLabel('overwrite existing')
        self.overwrite_cb = QCheckBox()

        layout.addWidget(self.overwirte_label, col, 0)
        layout.addWidget(self.overwrite_cb, col, 1, 1, 2)
        col += 1

        self.create_pushButton = QPushButton('create new hardware')
        layout.addWidget(self.create_pushButton, col, 0, 1, 3)
        col += 1

        self.info_box = QTextEdit('')
        layout.addWidget(self.info_box, col, 0, 2, 3)
        col += 1

    def set_com_types_options(self, options):
        self.com_types_combob.clear()
        self.com_types_combob.addItems(tuple(options))

    def model_text_changed(self, callback):
        self.model_le.textChanged.connect(callback)

    def pretty_name_text_changed(self, callback):
        self.pretty_name_le.textChanged.connect(callback)

    def guess_button_clicked(self, callback):
        self.guess_pretty_name_btn.clicked.connect(callback)

    def com_texted_changed(self, callback):
        self.com_types_combob.currentTextChanged.connect(callback)

    def create_clicked(self, callback):
        self.create_pushButton.clicked.connect(callback)

    def get_company(self) -> str:
        return self.company_le.text()

    def set_company(self, text):
        self.company_le.setText(text)

    def get_model(self) -> str:
        return self.model_le.text()

    def set_model(self, text):
        print(text)
        self.model_le.setText(text)

    def set_authors(self, authors):
        self.authors_le.setText(authors)

    def get_authors(self) -> str:
        return self.authors_le.text()

    def get_pretty_name(self) -> str:
        return self.pretty_name_le.text()

    def set_pretty_name(self, text):
        self.pretty_name_le.setText(text)

    def get_overwrite(self) -> bool:
        return self.overwrite_cb.isChecked()

    def set_overwrite(self, checked) -> None:
        self.overwrite_cb.setChecked(checked)

    def get_com_option(self) -> str:
        return self.com_types_combob.currentText()

    def set_info_text(self, text):
        self.info_box.setHtml(text)

    def append_info(self, text):
        self.info_box.setHtml(self.info_box.toHtml() + text)



from pyqtgraph.Qt import QtGui, QtCore
import numpy as np
import pyqtgraph as pg
from ScopeFoundry.graphics.zoomable_map.zoomable_map import ZoomableMapImageItem
import skimage.data
import matplotlib.pyplot as plt

pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')

app = QtGui.QApplication([])


graph_layout = pg.GraphicsLayoutWidget()

plot = pg.PlotItem()
graph_layout.addItem(plot)
plot.setAspectLocked(1)
vb  = plot.getViewBox()


zmi = ZoomableMapImageItem(plot_item=plot, 
                       image=skimage.data.immunohistochemistry().swapaxes(0,1), rect=(-1,-1,6,2),
                       tile_size=64, z_value=100)

zmi.setTransform(zmi.transform.rotate(10.))

M = np.identity(3)
M[0,0] = M[1,1] = 1/500
M[0,1] = 0.2/500
M[0,2] = 0.001/500
M[2,0] = 0
t = QtGui.QTransform()
t.setMatrix(*M.flat)
zmi.setTransform(t)


im = skimage.data.chelsea()
Nx,Ny,_ = im.shape
alpha = np.zeros((Nx,Ny,1),dtype=im.dtype) + 255
print(im.shape, alpha.shape, repr(im.dtype))
im = np.concatenate([im, alpha], axis=2)
fill = np.zeros( (1,1,4), dtype=im.dtype)
fill[0,0,3] = 0

zmi2 = ZoomableMapImageItem(plot_item=plot, image=im, rect=(3,-0.5,2,2), tile_size=64, z_value=200, fill=fill)

histlut2 = pg.HistogramLUTItem(image=zmi2 )
graph_layout.addItem(histlut2)


x = np.arange(-5, 5, 0.03)
y = np.arange(-5, 5, 0.03)

xx, yy = np.meshgrid(x, y)
z = np.sin(xx**2 + yy**2) / (xx**2 + yy**2)
z[::10,::10] = 1
print(z.shape, z.dtype)
colormap = plt.cm.get_cmap("RdBu")  # cm.get_cmap("CMRmap")
colormap._init()
lut = (colormap._lut * 255).view(np.ndarray)  # Convert matplotlib colormap from 0-1 to 0 -255 for Qt


zmi3 = ZoomableMapImageItem(plot_item=plot, image=z, rect=(3,3,4,4), 
                        tile_size=1024, z_value=300, fill=np.NaN, 
                        levels=(-1.1,1.1), lut=lut)
                        
                        
histlut = pg.HistogramLUTItem(image=zmi3)
graph_layout.addItem(histlut)

imitem3 = pg.ImageItem(z)
rect = (0,3,4,4)
imitem3.setRect(pg.QtCore.QRectF(*rect))
plot.addItem(imitem3)


plot.setTitle("Zoomable Map Test PyqtGraph")

graph_layout.show()



if __name__ == '__main__':
    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()




from pyqtgraph.Qt import QtGui, QtCore
import numpy as np
import pyqtgraph as pg
from ScopeFoundry.graphics.zoomable_map.zoomable_map import ZoomableMapImageItem
import skimage.data
import h5py

f = h5py.File("/Users/esbarnard/gdal_pyramid_test/210304_131725_survey_scan.h5", 'r')
M = f['measurement/survey_scan']
im = np.array(M['image_strips'][0])
rect = M['strip_rects'][0]

im = np.hstack([im,]*100)

app = QtGui.QApplication([])

plot = pg.PlotWidget()
plot.setAspectLocked(1)
vb  = plot.getViewBox()


zmi = ZoomableMapImageItem(plot_item=plot, image=im, rect=None)

plot.setTitle("Zoomable Map Test PyqtGraph")
plot.show()




if __name__ == '__main__':
    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



from qtpy.QtWidgets import QComboBox, QLineEdit
from typing_extensions import TypedDict

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem
from .helper_func import new_q_completer


ITEM_TYPE = "update-setting"
DESCRIPTION = "update a setting with value, a setting or __<iteration letter>"


class UpdateSettingKwargs(TypedDict):
    setting: str
    value: str


class UpdateSetting(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        v = self.kwargs["value"]
        try:
            v = self.app.get_lq(v).val
        except:
            pass
        if isinstance(v, str):
            if "__" in v:
                letter = v[v.find("__") + 2]
                v = self.measure.iter_values[letter]
        self.app.get_lq(self.kwargs["setting"]).update_value(v)


LE_TOOL_TIP = """value used to update. Can be a value, a setting, 
or '__<iteration letter>' """


class UpdateSettingEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = DESCRIPTION

    def __init__(self, measure, paths) -> None:
        self.paths = paths
        super().__init__(measure)

    def setup_ui(self):
        self.setting_cb = QComboBox()
        self.setting_cb.setEditable(True)
        self.setting_cb.addItems(self.paths)
        self.setting_cb.setToolTip("setting to update")
        self.completer = completer = new_q_completer(self.paths)
        self.setting_cb.setCompleter(completer)
        self.group_box.layout().addWidget(self.setting_cb)
        self.value_le = QLineEdit()
        completer = new_q_completer(
            self.paths + ["True", "False", "__A", "__B", "__C", "__D"]
        )
        self.value_le.setCompleter(completer)

        self.value_le.setToolTip(LE_TOOL_TIP)
        self.group_box.layout().addWidget(self.value_le)

    def get_kwargs(self) -> UpdateSettingKwargs:
        path = self.setting_cb.currentText()
        val = self.value_le.text()
        return {"setting": path, "value": val}

    def set_kwargs(self, **kwargs):
        self.setting_cb.setCurrentText(kwargs["setting"])
        self.value_le.setText(kwargs["value"])
        self.value_le.selectAll()
        self.value_le.setFocus()



from ScopeFoundry import BaseApp
from ScopeFoundry.ndarray_interactive import ArrayLQ_QTableModel
from qtpy import QtWidgets

import logging
logging.basicConfig(level='DEBUG')

class LQArrayTestApp(BaseApp):
    
    name = 'LQArrayTestApp'
    
    def __init__(self,argv):
        BaseApp.__init__(self,argv)
        
        self.settings.New('test_array', dtype=float, array=True,  ro=False, fmt="%1.2f",
                  initial=[[147, 111 , 100]])#,[208, 8 , 100],[218, 45 , 100],[345, 1500 , 100,],[517, 8 , 100],[772, 300, 100]])

        self.test_array_model = ArrayLQ_QTableModel(
            self.settings.test_array, 
            col_names=['Center','FWHM', 'Amplitude'])
        
        self.ui = QtWidgets.QScrollArea()
        self.ui.setWidgetResizable(True)
        self.ui.setLayout(QtWidgets.QVBoxLayout())
        
        self.tableView1 = QtWidgets.QTableView()
        self.tableView1.setModel(self.test_array_model)
        self.ui.layout().addWidget(self.tableView1)

        self.tableView2 = QtWidgets.QTableView()
        self.tableView2.setModel(self.test_array_model)
        self.ui.layout().addWidget(self.tableView2)


        self.test_array_model3 = ArrayLQ_QTableModel(self.settings.test_array )
        self.tableView3 = QtWidgets.QTableView()
        self.tableView3.horizontalHeader().hide()
        self.tableView3.setModel(self.test_array_model3)
        self.ui.layout().addWidget(self.tableView3)

        self.settings.New('test_array4', dtype=float, array=True,  ro=False, fmt="%1.2f",
                  initial=[1, 2 , 3])
        self.test_array_model4 = ArrayLQ_QTableModel(self.settings.test_array4 )
        self.tableView4 = QtWidgets.QTableView()
        self.tableView4.setModel(self.test_array_model4)
        self.ui.layout().addWidget(self.tableView4)

        self.test_array_model5 = ArrayLQ_QTableModel(self.settings.test_array4 , transpose=True)
        self.tableView5 = QtWidgets.QTableView()
        self.tableView5.setModel(self.test_array_model5)
        self.ui.layout().addWidget(self.tableView5)

        for dtype in [str, bool, int]:
            lq_name = 'test_array_{}'.format(dtype.__name__)
            lq = self.settings.New(lq_name, dtype=dtype, array=True,  ro=False,
                      initial=[0,1,"20",0])

            tableModel = ArrayLQ_QTableModel(lq, row_names= [dtype.__name__,], fmt=lq.fmt, transpose=True) 
            tableView = QtWidgets.QTableView()
            tableView.setModel(tableModel)
            self.ui.layout().addWidget(tableView)
            
        #  int does not round
        self.settings.New("test_int", dtype=int, initial=124)
        
        self.settings_ui = self.settings.New_UI()
        self.ui.layout().addWidget(self.settings_ui)
        
        
        self.treeWidget = QtWidgets.QTreeWidget()
        self.treeWidget.setColumnCount(2)
        self.ui.layout().addWidget(self.treeWidget)
        self.tree_item = QtWidgets.QTreeWidgetItem(self.treeWidget, ["asdf", 'jkl;'])
        self.settings.add_widgets_to_subtree(self.tree_item)
        
        
        self.load_ini_button = QtWidgets.QPushButton("Load INI file")
        self.ui.layout().addWidget(self.load_ini_button)
        
        self.load_ini_button.clicked.connect(self.on_load_ini_button)
        

        self.ui.show()
        self.setup_console_widget().show()
        
    def on_load_ini_button(self):
        self.settings_load_ini('lq_array_test.ini')
        
        
if __name__ == '__main__':
    app = LQArrayTestApp([])
    app.exec_()


from qtpy.QtWidgets import QLabel
from typing_extensions import TypedDict

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem

ITEMTYPE = "pause"


class PauseKwargs(TypedDict):
    info: str


class Pause(BaseItem):
    item_type = ITEMTYPE

    def visit(self):
        self.measure.settings["paused"] = True




class PauseEditorUI(EditorBaseUI):
    item_type = ITEMTYPE

    description = "pauses - click resume"

    def setup_ui(self):
        self.pause_spacer = QLabel()
        self.group_box.layout().addWidget(self.pause_spacer)

    def get_kwargs(self) -> PauseKwargs:
        return {"info": "click resume to continue"}

    def set_kwargs(self, **kwargs):
        self.pause_spacer.setFocus()



"""
author: Camille Stavrakas, Edward Barnard, Benedikt Ursprung
"""
import datetime
import time
from typing import Sequence

import numpy as np
import pyqtgraph as pg
from qtpy import QtWidgets
from ScopeFoundry import Measurement, h5_io

DESCRIPTION = """
<p>Sweeps a setting z within a range, measures an optimization quantity f(z)
and calculates z0 such that:</p>
<p>f(z0) >= f(z) for all z</p>
<p>finally sets z = z0 + z_offset</p>
"""


def mk_data_dict(center, span, num, z_original):
    return {
        "z_original": z_original,
        "z0_coarse": center,
        "z0_fine": center,
        "z0": center,
        "z_coarse": mk_z_array(center, span, num),
        "f_coarse": np.zeros(num),
        "z_fine": mk_z_array(center, span, 2 * num),
        "f_fine": np.zeros(2 * num),
    }


SAME_AS_Z = "same_as_z"


class RangedOptimization(Measurement):
    name = "ranged_optimization"

    def __init__(
        self,
        app,
        name=None,
        lq_kwargs={"spinbox_decimals": 6, "unit": ""},
        range_initials=[0, 10, 0.1],
    ):
        self.range_initials = range_initials
        self.lq_kwargs = lq_kwargs
        self.z0_history = []
        self.f_history = []
        self.z_history = []
        self.time_history = []
        self.z_read_path = ""
        Measurement.__init__(self, app, name)

    def setup(self):
        self.settings.New("f", str, choices=("",), description=DESCRIPTION)
        self.settings.New("N_samples", int, initial=10)
        self.settings.New(
            "sampling_period",
            float,
            initial=0.050,
            unit="s",
            si=True,
            description="time waited between sampling",
        )
        self.settings.New(
            "use_current_z_as_center",
            bool,
            initial=True,
            description="instead of <b>z_center</b> the current value of <b>z</b> is used",
        )
        self.settings.New_Range(
            "z",
            include_center_span=True,
            initials=self.range_initials,
            description="defines a range over which <b>z</b> is varied",
            **self.lq_kwargs,
        )
        self.settings.New(
            "z_offset",
            initial=0,
            description="an offset that will be applied when moving to optimal <i>z</i> value after optimization",
            **self.lq_kwargs,
        )
        self.settings.New(
            "use_fine_optimization",
            bool,
            initial=True,
            description="optimization runs again around z0 from first run",
        )
        self.settings.New("coarse_to_fine_span_ratio", initial=4.0)
        self.settings.New("z_span_travel_time", initial=2.0, unit="sec")
        self.settings.New(
            "z",
            str,
            choices=[],
            description="path to a setting that can manipulate the sweep value z",
        )
        self.settings.New(
            "z_read",
            str,
            initial=SAME_AS_Z,
            choices=[SAME_AS_Z],
            description=f"if not <i>{SAME_AS_Z}</i> this setting will be used to get actual value of z",
        )
        cs = list(post_processors.keys())
        self.settings.New(
            "post_processor",
            str,
            initial="gauss_mean",
            choices=cs,
            description="e.g. fit gaussian to data and use the derived mean as optimized value",
        )
        self.settings.New("take_post_process_value", bool, initial=False)
        self.settings.New("save_h5", bool, initial=True)

        s = self.settings
        self.data = mk_data_dict(s["z_center"], s["z_span"], s["z_num"], 0)
        self.add_operation("filter choices", self.filter_choices)
        self.add_operation("unfilter choices", self.unfilter_choices)
        self.add_operation("post process", self.post_process)
        self.add_operation("go to optimal value", self.go_to_optimal_value)
        self.add_operation("go to post process value",
                           self.go_to_post_process_value)

    def setup_figure(self):
        s = self.settings

        settings_layout = QtWidgets.QHBoxLayout()
        settings_layout.addWidget(
            s.New_UI(("z_span", "z_num", "use_current_z_as_center", "z_center")))
        settings_layout.addWidget(s.New_UI(("f", "z", "z_read", "z_offset")))
        settings_layout.addWidget(s.New_UI(
            ("N_samples", "sampling_period", "use_fine_optimization", "post_processor")))

        take_pushButton = QtWidgets.QPushButton("interrupt and take")
        take_pushButton.clicked.connect(self.take_current_optimal)
        pp_btn = QtWidgets.QPushButton("post process")
        pp_btn.clicked.connect(self.post_process)
        go_to_pp_btn = QtWidgets.QPushButton("go to processed value")
        go_to_pp_btn.clicked.connect(self.go_to_post_process_value)
        btns_layout = QtWidgets.QHBoxLayout()
        btns_layout.addWidget(s.get_lq("activation").new_pushButton())
        btns_layout.addWidget(take_pushButton)
        btns_layout.addWidget(pp_btn)
        btns_layout.addWidget(go_to_pp_btn)

        self.ui = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(settings_layout)
        layout.addLayout(btns_layout)
        layout.addWidget(self.mk_graph_widget())
        self.ui.setLayout(layout)

        self.unfilter_choices()

    def run(self):
        s = self.settings

        self._take_current_flag = False
        self.set_post_process_lines_visible(False)
        self.update_z_read()
        self.try_connect_hws()
        self.update_z_center()
        z_ori = self.read_z()
        d = self.data = mk_data_dict(
            s["z_center"], s["z_span"], s["z_num"], z_ori)

        self.ii = 0
        self.total_dpt = (1 + 2 * s["use_fine_optimization"]) * s["z_num"]

        # coarse sweep
        self.sweep("coarse")

        # fine sweep
        if s["use_fine_optimization"] and not self.interrupt_measurement_called:
            r = self.settings["coarse_to_fine_span_ratio"]
            d["z_fine"] = mk_z_array(d["z0_coarse"], s["z_span"] / r, 2 * s["z_num"])
            d["f_fine"] = np.ones_like(d["z_fine"]) * max(d["f_coarse"])
            self.sweep("fine")

        # go to final position
        if self.interrupt_measurement_called:
            if self._take_current_flag:
                val = d["z0"] + s["z_offset"]
                msg = f"{self.name} interrupted with take, moving optimal to {val:0.2f}"
            else:
                val = d['z_original']
                msg = f"{self.name} interrupted, moving to original position to {val:0.2f}"
            timeout = timeout = s["z_span_travel_time"] / 4
            self.write_z_target(val, timeout, msg)
        else:
            if s["take_post_process_value"]:
                self.go_to_post_process_value()
            else:
                self.go_to_optimal_value()
            self.update_history(d["z0"])
            self.save_h5()
            self.print_history()
            
    def sweep(self, name="coarse"):
        s = self.settings
        d = self.data
        
        self.write_z_target(d[f"z_{name}"][0], s["z_span_travel_time"])
        for i, z in enumerate(d[f"z_{name}"]):
            self.set_progress((self.ii + 1) * 100 / self.total_dpt)
            self.write_z_target(
                z, s["z_span_travel_time"] / s["z_num"] / 4)
            d[f"f_{name}"][i] = self.read_f()
            d["z0"] = d[f"z0_{name}"] = d[f"z_{name}"][np.argmax(d[f"f_{name}"])]
            if self.interrupt_measurement_called:
                break
            self.ii += 1
            
    def update_display(self):
        s = self.settings
        d = self.data

        self.plot_line_coarse.setData(d["z_coarse"], d["f_coarse"])
        self.plot_line_fine.setData(d["z_fine"], d["f_fine"])

        self.line_z_original.setPos(d["z_original"])
        self.line_z0_coarse.setPos(d["z0_coarse"])
        self.line_z0_fine.setPos(d["z0_fine"])

        self.line_z0_fine.setVisible(s["use_fine_optimization"])
        self.plot_line_fine.setVisible(s["use_fine_optimization"])

    def save_h5(self):
        h5_file = h5_io.h5_base_file(app=self.app, measurement=self)
        group = h5_io.h5_create_measurement_group(self, h5_file)
        self.data["times_history"] = [
            int(t.timestamp()) for t in self.time_history]
        self.data["z0_history"] = self.z0_history
        for k, v in self.data.items():
            group[k] = np.array(v)
        h5_file.close()

    def update_history(self, z0):
        s = self.settings
        self.z0_history.append(z0)
        self.f_history.append(s["f"])
        self.z_history.append(s["z_read"])
        self.time_history.append(datetime.datetime.now())

    def print_history(self):
        template = "{0:%y%m%d_%H%M%S}: {2:35} {1:8.4}"
        for rec in zip(self.time_history, self.z0_history, self.f_history):
            print(template.format(*rec))

    def read_f(self):
        x = 0.0
        for _ in range(self.settings["N_samples"]):
            x += self.app.read_setting(self.settings["f"])
            time.sleep(self.settings["sampling_period"])

        return x / self.settings["N_samples"]

    def write_z_target(self, value, timeout=0, msg=None):
        self.app.write_setting(self.settings["z"], value)
        if msg:
            print(msg)
        time.sleep(timeout)

    def read_z(self):
        return self.app.read_setting(self.z_read_path)

    def update_z_read(self):
        s = self.settings
        if s["z_read"] == SAME_AS_Z:
            self.z_read_path = s["z"]
        else:
            self.z_read_path = s["z_read"]

    def update_z_center(self):
        s = self.settings
        if s["use_current_z_as_center"]:
            s["z_center"] = self.read_z() - s["z_offset"]
            
    def try_connect_hws(self):
        """in case user forgot to connect to corresponding hardware"""
        self.connect_hw(self.settings["z"])
        self.connect_hw(self.settings["z_read"])
        self.connect_hw(self.settings["f"])

    def connect_hw(self, lq_path):
        if not lq_path.startswith("hw"):
            return
        try:
            sec, hw_name, _ = lq_path.split("/")
            print(self.name, "setting",
                  "/".join([sec, hw_name, "connected"]), "to True")
            self.app.write_setting("/".join([sec, hw_name, "connected"]), True)
        except (ValueError, KeyError) as e:
            print(self.name, "could connect")
            print(e)

    def take_current_optimal(self):
        self._take_current_flag = True
        self.interrupt_measurement_called = True

    def post_process(self):
        """apply an algorithm to find a derived optimized quantity"""
        s = self.settings
        if s["use_fine_optimization"]:
            z, f = self.data["z_fine"], self.data["f_fine"]
        else:
            z, f = self.data["z_coarse"], self.data["f_coarse"]

        z0, fs = post_process(z, f, s["post_processor"])

        self.set_post_process_lines_visible(True)
        self.plot_line_post_process.setData(z, fs)
        self.line_z0_post_process.setPos(z0)
        return z0

    def set_post_process_lines_visible(self, visible: bool):
        self.plot_line_post_process.setVisible(visible)
        self.line_z0_post_process.setVisible(visible)

    def go_to_post_process_value(self):
        s = self.settings
        val = self.post_process() + s["z_offset"]
        msg = f"{self.name} going to post process position {val:0.2f}"
        self.write_z_target(val, msg=msg)

    def go_to_optimal_value(self):
        s = self.settings
        val = self.data["z0"] + s["z_offset"]
        msg = f"{self.name} going to optimal position {val:0.2f}"
        self.write_z_target(val, msg=msg)

    def _set_choices(self, name, choices):
        before = self.settings[name]
        self.settings.get_lq(name).change_choice_list(choices)
        if not before in choices:
            before = choices[0]
        self.settings[name] = before

    def filter_choices(self):
        self._set_choices("f", self.app.get_setting_paths(True, False))
        self._set_choices("z", self.app.get_setting_paths(False, True))
        self._set_choices("z_read", [SAME_AS_Z] + 
                          self.app.get_setting_paths(True, False))

    def unfilter_choices(self):
        choices = self.app.get_setting_paths()
        self._set_choices("f", choices)
        self._set_choices("z", choices)
        self._set_choices("z_read", [SAME_AS_Z] + choices)

    def mk_graph_widget(self):
        widget = pg.GraphicsLayoutWidget(border=(0, 0, 0))
        axes = widget.addPlot()
        axes.showGrid(x=True, y=True, alpha=0.3)

        self.plot_line_coarse = axes.plot()
        self.plot_line_fine = axes.plot(pen="r")
        self.plot_line_post_process = axes.plot(pen="g")

        # indicator lines
        self.line_z_original = pg.InfiniteLine(
            movable=False,
            pen="b",
            label="original position: {value:0.6f}",
            labelOpts={
                "color": "b",
                "movable": True,
                "position": 0.15,
                "fill": (200, 200, 200, 200),
            },
        )
        self.line_z0_coarse = pg.InfiniteLine(
            movable=False,
            pen=(200, 200, 200),
            label="coarse optimized: {value:0.6f}",
            labelOpts={
                "color": (200, 200, 200),
                "movable": True,
                "position": 0.30,
                "fill": (200, 200, 200, 60),
            },
        )
        self.line_z0_fine = pg.InfiniteLine(
            movable=False,
            pen="r",
            label="fine optimized: {value:0.6f}",
            labelOpts={
                "color": "r",
                "movable": True,
                "position": 0.45,
                "fill": (200, 200, 200, 80),
            },
        )
        self.line_z0_post_process = pg.InfiniteLine(
            movable=False,
            pen="g",
            label="post process value: {value:0.6f}",
            labelOpts={
                "color": "g",
                "movable": True,
                "position": 0.50,
                "fill": (200, 200, 200, 80),
            },
        )
        axes.addItem(self.line_z_original, ignoreBounds=True)
        axes.addItem(self.line_z0_coarse, ignoreBounds=True)
        axes.addItem(self.line_z0_fine, ignoreBounds=True)
        axes.addItem(self.line_z0_post_process, ignoreBounds=True)
        axes.setLabels(bottom="z", left="f")
        axes.enableAutoRange()

        return widget


def mk_z_array(center, span, num):
    return np.linspace(center - 0.5 * span, center + 0.5 * span, num)


def gauss(x, a, x0, sigma):
    return a * np.exp(-((x - x0) ** 2) / (2 * sigma ** 2))


def fit_gauss(x, y):
    from scipy.optimize import curve_fit

    p0 = [y.max(), x[y.argmax()], 10]
    popt, _ = curve_fit(gauss, x, y, p0=p0)
    return popt  # a,mean,sigma


def post_process_gauss(z, f):
    a, mean, sigma = fit_gauss(z, f - f.min())
    f0 = gauss(z, a, mean, sigma) + f.min()
    z0 = mean
    return z0, f0


def post_process_min(z, f):
    return z[np.array(f.argmin())], f


def post_process_max(z, f):
    return z[np.array(f.argmax())], f


post_processors = {
    "gauss_mean": post_process_gauss,
    "min": post_process_min,
    "max": post_process_max,
}


def post_process(x: Sequence, y: Sequence, post_processor: str="gauss_mean"):
    func = post_processors[post_processor]
    return func(x, y)



from pathlib import Path

from qtpy import QtWidgets

from ScopeFoundry.data_browser import DataBrowserView


class FileInfoView(DataBrowserView):
    """A general viewer to handle text files and
    unsupported file types."""

    name = "file_info"

    def setup(self):
        self.ui = QtWidgets.QTextEdit("file_info")

    def on_change_data_filename(self, fname=None):
        if fname is None:
            fname = self.databrowser.settings["data_filename"]

        # Use pathlib
        fname = Path(fname)

        ext = fname.suffix

        if ext in (".py", ".ini", ".txt", ".yml", ".yaml"):
            with open(fname, "r") as f:
                self.ui.setText(f.read())
        else:
            self.ui.setText(str(fname))

    def is_file_supported(self, fname):
        return True



import datetime
from enum import Enum
from functools import partial
from pathlib import Path
from re import sub
from string import Template


class ComTypes(Enum):
    OTHER = 'OTHER'
    DLL = 'DLL'
    SERIAL = 'SERIAL'
    # VISA = 'VISA'


def to_class_name(s):
    s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return ''.join([s[0].upper(), s[1:]])


def do_template_file(template_file, target_file_name, substitutions, path_to_module):
    with open(template_file, 'r') as f:
        template = Template(f.read())

    with open(path_to_module / Path(target_file_name), 'w') as f:
        f.write(template.substitute(**substitutions))


def new_hw(company='thorlabs',
           model='XX6000_glavo_mirror',
           authors='Benedikt Ursprung, Mark E.Ziffer',
           pretty_name='galvo_mirror',
           overwrite_existing_module=True,
           com_type=ComTypes.DLL):

    module_name = f"{company.lower()}_{model.lower()}"
    path_to_module = Path('../../ScopeFoundryHW') / Path(module_name)

    if path_to_module.exists() and not overwrite_existing_module:
        raise FileExistsError(
            f'{module_name} already exists! You can use overwrite flag at your own risk')

    today = datetime.datetime.today()
    pretty_date = today.strftime("%b %d, %Y")
    year = today.strftime("%Y")

    if pretty_name is None or pretty_name == "":
        pretty_name = (model.lower())
    pretty_name = pretty_name.lower()

    dev_file_name = pretty_name + '_dev.py'
    dev_class_name = to_class_name(model.upper() + '_DEV')
    import_dev = f"from .{dev_file_name.rstrip('.py')} import {dev_class_name}"

    hw_file_name = pretty_name.lower() + '_hw.py'
    hw_class_name = to_class_name(model.upper() + '_HW')

    readout_file_name = pretty_name.lower() + '_readout.py'
    readout_class_name = to_class_name(model.upper() + '_Readout')

    test_app_file_name = pretty_name.lower() + '_test_app.py'

    print(hw_file_name, hw_file_name.rstrip('.py'))

    init_imports = f"from .{hw_file_name.rstrip('.py')} import {hw_class_name} \nfrom .{readout_file_name.rstrip('.py')} import {readout_class_name}"
    test_app_imports = f"from ScopeFoundryHW.{module_name.rstrip('.py')} import {hw_class_name}, {readout_class_name}"

    print(init_imports)

    substitutions = {
        'DATE_PRETTY': pretty_date,
        'YEAR': year,
        'AUTHORS': authors,
        'AUTHORS_SPLIT_NEWLINE': '\n'.join(authors.split(', ')),
        'DEV_CLASS_NAME': dev_class_name,
        'IMPORT_DEV': import_dev,
        'HW_CLASS_NAME': hw_class_name,
        'READOUT_CLASS_NAME': readout_class_name,
        'READOUT_NAME': pretty_name.lower() + '_readout',
        'TEST_APP_NAME': pretty_name.lower() + '_test_app',
        'HW_NAME': pretty_name.lower(),
        'INIT_IMPORTS': init_imports,
        'IMPORTS_IN_TEST_APP': test_app_imports,
        'MODULE_NAME': module_name,
        'COMPANY': " ".join(company.split('_')),
        'MODEL': " ".join(model.split('_'))
    }

    path_to_module.mkdir(exist_ok=True)
    do = partial(do_template_file,
                 substitutions=substitutions,
                 path_to_module=path_to_module
                 )

    do('templates/_readout.py', readout_file_name)
    do('templates/_test_app.py', test_app_file_name)
    do('templates/___init__.py', '__init__.py')
    do('templates/_LICENSE', 'LICENSE')
    do('templates/_README.md', 'README.md')

    if com_type == ComTypes.DLL:
        do('templates/_dll_dev.py', dev_file_name)
        do('templates/_dll_hw.py', hw_file_name)
    if com_type == ComTypes.SERIAL:
        do('templates/_serial_dev.py', dev_file_name)
        do('templates/_serial_hw.py', hw_file_name)
    if com_type == ComTypes.OTHER:
        do('templates/_serial_dev.py', dev_file_name)
        do('templates/_serial_hw.py', hw_file_name)

    return module_name


if __name__ == '__main__':
    new_hw(com_type=ComTypes.SERIAL)



import operator

from qtpy.QtWidgets import QComboBox, QLineEdit
from typing_extensions import TypedDict

from .helper_func import new_q_completer
from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem

ITEM_TYPE = "interrupt-if"


class InterruptIfKwargs(TypedDict):
    setting: str
    operator: str
    value: str


class InterruptIf(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        relate = {"=": operator.eq, ">": operator.gt, "<": operator.lt}[
            self.kwargs["operator"]
        ]
        lq = self.app.get_lq(self.kwargs["setting"])
        val = lq.coerce_to_type(self.kwargs["value"])
        if relate(lq.val, val):
            self.measure.interrupt()


class IterruptIfEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = "interrupt if a condition is met"

    def __init__(self, measure, paths) -> None:
        self.paths = paths
        super().__init__(measure)

    def setup_ui(self):
        self.setting_cb = QComboBox()
        self.setting_cb.setEditable(True)
        self.setting_cb.addItems(self.paths)
        self.setting_cb.setToolTip("setting")
        completer = new_q_completer(self.paths)
        self.setting_cb.setCompleter(completer)
        self.layout.addWidget(self.setting_cb)
        self.operator_cb = QComboBox()
        self.operator_cb.addItems(["=", "<", ">"])
        self.layout.addWidget(self.operator_cb)
        self.value_le = QLineEdit()
        self.value_le.setCompleter(completer)
        self.layout.addWidget(self.value_le)

    def get_kwargs(self) -> InterruptIfKwargs:
        path = self.setting_cb.currentText()
        o = self.operator_cb.currentText()
        val = self.value_le.text()
        return {"setting": path, "operator": o, "value": val}

    def set_kwargs(self, **kwargs):
        self.setting_cb.setEditText(kwargs["setting"])
        self.operator_cb.setEditText(kwargs["operator"])
        self.value_le.setText(kwargs["value"])



from pyqtgraph.Qt import QtGui, QtCore
import numpy as np
import pyqtgraph as pg
import skimage.data
import math

def printTransform(t):
    print(f"[ {t.m11()} {t.m12()} {t.m13()} ]")
    print(f"[ {t.m21()} {t.m22()} {t.m23()} ]")
    print(f"[ {t.m31()} {t.m32()} {t.m33()} ]")



pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')

app = QtGui.QApplication([])
plot = pg.PlotWidget()
plot.setAspectLocked(1)
vb  = plot.getViewBox()
plot.show()

x = np.arange(-5, 5, 0.1)
y = np.arange(-5, 5, 0.1)

xx, yy = np.meshgrid(x, y)
sin_im = np.sin(xx**2 + yy**2) / (xx**2 + yy**2)
sin_im = skimage.data.camera()[:500,:500]
#sin_im = np.random.random(sin_im.shape)
imitem = pg.ImageItem(sin_im)
rect = (0,3,4,4)
imitem.setRect(pg.QtCore.QRectF(*rect))
#imitem.setPos()

#imitem.resetTransform()
t = imitem.transform()
#t.set
printTransform(t)

M = np.identity(3)
M[0,0] = M[1,1] = 3
M[0,1] = 0.2
M[0,2] = 0.001
M[2,0] = -1000
t.setMatrix(*M.flat)
imitem.setTransform(t)


t = QtGui.QTransform()
t.scale(1/512., 1/512.)
t.rotate(5.)
imitem.setTransform(t)


    
printTransform(t)

print("transform", imitem.transform())
print("transform", imitem.transform())
plot.addItem(imitem)


#x0, y0, w, h = self.zmi.rect
#0,0, 1, 1

#orig_pixel_size_x = 1 / zmi.image.shape[0]
#orig_pixel_size_y = 1 / zmi.image.shape[1]
#print("orig_pixel_size", orig_pixel_size_x, orig_pixel_size_y)


def tile_transform(im, tile_size, z,ii,jj):
    zf = 2**z
    
    Nx, Ny = im.shape
    
    
    Nt = tile_size 
    
    x = ii*zf* Nt
    y = jj*zf* Nt

    tr = QtGui.QTransform()
    
    sx = zf#*Nt/Nx * Nt
    sy = zf#*Nt/Ny * Nt
    tr.translate(x, y)
    tr.scale(sx, sy)
    
    

    print(f"Nt={Nt}, z={z}, zf={zf}, ii={ii}, jj={jj}, x={x}, y={y}, scale {sx} {sy}")

    return tr

Nt = 32

for (zoom,ii,jj) in [(0,0,0),(0,1,1),(1,1,1),(2,1,1),(3,1,1)]:#, (1,1,1),(1,0,1)]:
    t_img = np.random.random((Nt,Nt))
    t_imitem = pg.ImageItem(t_img)
    tr = tile_transform(sin_im, Nt,zoom,ii,jj)*imitem.transform()
    printTransform(tr)
    t_imitem.setTransform(tr)
    plot.addItem(t_imitem)


def on_range_changed(src, new_range=None):
    if new_range is None:
            new_range = vb.viewRange()
    (xl,xr), (yb, yt) = new_range

    print('on_range_changed', new_range)
    print('vb.viewPixelSize()', vb.viewPixelSize())
    tI = imitem.transform().inverted()[0]
    print('converted_new_range BL', tI.map(xl,yb))
    print('converted_new_range TR', tI.map(xr,yt))
    #print('viewPixelSize --> orig', t.inverted()[0].map(*vb.viewPixelSize()))
    #print('vb.viewRect', vb.viewRect())
    
    px, py = vb.viewPixelSize()
    
    #distances_in_orig_pixel_space
    x0, y0 = tI.map(xl,yb)
    x1, y1 = tI.map(xr,yt)
    
    bl_tr_dist_view = np.sqrt( (xr-xl)**2 + (yt-yb)**2)
    bl_tr_dist_view_px = np.sqrt( ((xr-xl)/px)**2 + ((yt-yb)/py)**2)
    bl_tr_dist_orig = np.sqrt( (x1-x0)**2 + (y1-y0)**2)

    print("bl_tr_dist_view", bl_tr_dist_view)
    print("bl_tr_dist_view_px", bl_tr_dist_view_px)
    print("bl_tr_dist_orig", bl_tr_dist_orig)
    
        
    screen_pixels_in_view_x = abs(xr-xl)/px
    screen_pixels_in_view_y = abs(yt-yb)/py
    
    print('screen_pixels_in_view', screen_pixels_in_view_x, screen_pixels_in_view_y)
    
    #distances_in_orig_pixel_space
    x0, y0 = tI.map(xl,yb)
    x1, y1 = tI.map(xr,yt)
    print("distances_in_orig_pixel_space", x1-x0, y1-y0)
    
    print("approx zoom factors", screen_pixels_in_view_x/(x1-x0), screen_pixels_in_view_x/(y1-y0))
    zf = bl_tr_dist_orig/bl_tr_dist_view_px
    print("zoom factor", zf)
    
    z_max = 10 # temporary
    zoom = min(max(int(np.floor(np.log2(zf))),0), z_max)
    zfi = 2**zoom # integer power of two zoom factor    

    
    # tile position for a given zoom in original pixel coords
    #x = ii*zf* Nt
    #y = jj*zf* Nt
    # flipping this formula
    #ii = x / (zf*Nt)
    #jj = y / (zf*Nt)

    Nt = 32
    
    Nx_tiles = int(np.ceil(sin_im.shape[0]/zfi/Nt))
    Ny_tiles = int(np.ceil(sin_im.shape[1]/zfi/Nt))
    
    print("zoom", zoom,'zfi', zfi, "Tiles at zoom", Nx_tiles, Ny_tiles )
    
    
    ii0 = x0 / (zf*Nt)
    ii0 = max(0, math.floor(ii0)) # clip at zero
    ii1 = x1 / (zf*Nt)
    ii1 = min(math.ceil(ii1), Nx_tiles)

    jj0 = y0 / (zf*Nt)
    jj0 = max(0, math.floor(jj0)) # clip at zero
    jj1 = y1 / (zf*Nt)
    jj1 = min(math.ceil(jj1), Ny_tiles)    
    
    
    print(f"x tiles {Nt} within view {ii0} to {ii1}")
    print(f"y tiles {Nt} within view {jj0} to {jj1}")
    
    def tiles_at_zoom(z):
        zfi = 2**z
        Nt = tile_size = 32
        #Nx_tiles = int(np.ceil(self.image.shape[0]/zfi/Nt))
        #Ny_tiles = int(np.ceil(self.image.shape[1]/zfi/Nt))
        Nx_tiles = (sin_im.shape[0]/zfi/Nt)
        Ny_tiles = (sin_im.shape[1]/zfi/Nt)
        
        return (Nx_tiles, Ny_tiles)
        
    print("tiles_at_zoom", zoom, tiles_at_zoom(zoom))
    
sigprox = pg.SignalProxy(
                    signal=plot.sigRangeChanged,
                    #delay=0.1,
                    rateLimit=30,
                    slot=on_range_changed)


if __name__ == '__main__':
    import sys
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



'''
Created on Jul 20, 2018

@author: Benedikt Ursprung
'''

from qtpy import  QtCore, QtWidgets, QtGui
import pyqtgraph as pg
import numpy as np

class MinMaxQSlider(QtWidgets.QWidget):
    """
    Costume QSlider widget designed to work with logged_quantities.
    Two spin boxes allow to dynamically update the min and max value of the slider.
    """
    
    #Signals
    updated_value = QtCore.Signal((float,),(int,))
    
    def __init__(self, name="", spinbox_decimals = 2, **QSlider_kwargs):

        self.name = name
        
        #Setup widget
        QtWidgets.QWidget.__init__(self)
        
        #slider
        self.slider = QtWidgets.QSlider(orientation = QtCore.Qt.Horizontal, **QSlider_kwargs)
        self.slider.valueChanged.connect(self.on_slider_value_changed)
        self.slider.setSingleStep(1)
        self.slider.sliderMoved[int].connect(self.set_title)
        
        #min and max spinbox
        self.min_double_spinbox = QtWidgets.QDoubleSpinBox()
        self.min_double_spinbox.setRange(-1e12, 1e12)
        self.max_double_spinbox = QtWidgets.QDoubleSpinBox()
        self.max_double_spinbox.setRange(-1e12, 1e12)
        
        self.min_double_spinbox.valueChanged[float].connect(self.update_slider_minimum)
        self.max_double_spinbox.valueChanged[float].connect(self.update_slider_maximum)
        
        self.setDecimals(spinbox_decimals)

        #title label
        self.title_label = QtWidgets.QLabel()
        
        #Setup layout
        self.layout_ = QtWidgets.QGridLayout()
        self.layout_.addWidget(self.min_double_spinbox, 0,0, alignment=QtCore.Qt.AlignLeft)
        self.layout_.addWidget(self.title_label, 0,1, alignment = QtCore.Qt.AlignHCenter)
        self.layout_.addWidget(self.max_double_spinbox, 0,2, alignment=QtCore.Qt.AlignRight)    
        self.layout_.addWidget(self.slider, 1,0, 1,3)
        
        self.setLayout(self.layout_)


    def set_name(self,name):
        self.name = name
        val = self.transform_from_slider(self.slider.value())
        self.set_title(val)
        
    @property
    def vmin(self):
        return self.min_double_spinbox.value()
    @property
    def vmax(self):
        return self.max_double_spinbox.value()
    @property
    def vrange(self):
        return self.vmax-self.vmin
        
    def transform_to_slider(self,x):
        pct = 100*(x-self.vmin)/self.vrange
        return int(pct)
    
    def transform_from_slider(self,x):
        val = self.vmin + (x*self.vrange/100)
        return val
        
    def update_min_double_spinbox(self, vmin):
        try:
            self.min_double_spinbox.blockSignals(True)
            self.min_double_spinbox.setValue(vmin)
        finally:
            self.min_double_spinbox.blockSignals(False)   
                     
    def update_max_double_spinbox(self, vmax):        
        try:
            self.max_double_spinbox.blockSignals(True)
            self.max_double_spinbox.setValue(vmax)
        finally:
            self.max_double_spinbox.blockSignals(False)
        
    def update_slider_minimum(self, vmin):
        val = self.slider.sliderPosition()
        self.slider.setMinimum(self.transform_to_slider(vmin))
        self.slider.setSliderPosition(val)
    def update_slider_maximum(self, vmax):
        self.slider.setMaximum(self.transform_to_slider(vmax))
        
    def update_value(self,x):
        try:
            self.slider.blockSignals(True)
            self.slider.setValue(self.transform_to_slider(x))
        finally:
            self.slider.blockSignals(False) 
        
    def on_slider_value_changed(self):
        val = self.transform_from_slider(self.slider.value())
        self.updated_value.emit(val)

    def setRange(self, vmin, vmax):
        # as it is currently used in LoggedQuantity.change_min_max()
        self.update_min_double_spinbox(vmin)
        self.update_max_double_spinbox(vmax)
        
    def setDecimals(self,decimals):
        self.min_double_spinbox.setDecimals(decimals)
        self.max_double_spinbox.setDecimals(decimals)
        
    def setSuffix(self,unit):
        self.min_double_spinbox.setSuffix(" "+unit)
        self.max_double_spinbox.setSuffix(" "+unit)

    def setSingleStep(self, spinbox_step):
        self.min_double_spinbox.setSingleStep(spinbox_step)
        self.max_double_spinbox.setSingleStep(spinbox_step)
        
    def set_title(self, val):
        val = self.transform_from_slider(val)
        text =self.name+" "+str(val)
        self.title_label.setText(text)

        
class RegionSlicer(QtWidgets.QWidget):
    '''
    **Bases:** :class: `QWidget <pyqt.QtWidgets.QWidget>`
    
    Adds a movable <pyqtgraph.LinearRegionItem> to a plot.
    Provides a numpy slice and mask that would slices the x_array according to the region.
    
    ===============================  =============================================================================
    **Signals:**
    region_changed_signal()          Emitted when region is changed or activated by user
    ===============================  =============================================================================
    
    
    ===============================  =============================================================================
    **Properties:**
    slice:                           numpy.slice[start:stop]
    s_:                              :property:slice if activated  else *s_return_on_deactivated*
    mask:                            numpy boolean array of size 'x_array': True for where x is within region
                                     false otherwise (ignores step). Useful if x_array is not ordered. 
    m_:                              mask if activated, else a mask equivalent to 
                                        *s_return_on_deactivated*
    ===============================  =============================================================================    
    '''
    
    region_changed_signal = QtCore.Signal()
    
    def __init__(self, plot_item, x_array=None, name='array_slicer_name',
                 slicer_updated_func=lambda:None,  
                 initial=[0,100], s_return_on_deactivated = np.s_[:],
                 brush=QtGui.QColor(0,255,0,70), ZValue=10, 
                 font=QtGui.QFont("Times", 12), label_line=1, activated=False):
        """Create a new LinearRegionItem on plot_item.
        
        ====================== ==============================================================
        **Arguments:**
        plot_item              <pyqtgraph.PlotDataItem> (recommended) 
                                or <pyqtgraph.PlotItem>  (does not grab x_array data from plot
                                item: initialize x_array manually
                                or use :func:`set_x_array <self.set_x_array>`)
        x_array                use to initialize x_array if plot_item == <pyqtgraph.PlotItem>.
        name                   <str> 
        slicer_updated_func    gets called when region is updated, alternatively use 
                               :sig:region_changed_signal().
        initial                initial index values [start,stop,step] or [start,stop]
        s_return_on_deactivated  Object returned if RegionSlicer is not activated.
                                 Slicing with np.s_[:] (default) and np.s_[0] gives the full and 
                                 an empty array respectively. Note, <RegionSlicer.slice> always 
                                 returns the last determined slice even if slice is deactivated.
        brush,ZValue,          are passed to the LinearRegionItem
        font,                  passed to the label
        label_line             '0' or '1' for placement of label onto 'left' or 'right'
                               bounding line. 
        activated              <bool> state at initialization
        ====================== ===============================================================
        """
        QtWidgets.QWidget.__init__(self)        

        self.name = name
        
        self.slicer_updated = slicer_updated_func
        
        from ScopeFoundry.logged_quantity import LQCollection
        self.settings = LQCollection()
        if len(initial)==2: initial.append(1)
        start,stop,step = initial     
        self.start = self.settings.New('start', int, initial=start, vmin = 0)
        self.stop = self.settings.New('stop', int, initial=stop, vmin = 0)
        self.step = self.settings.New('step', int, initial=step)
        self.activated = self.settings.New('activated', bool, initial = activated)
        
        self.s_return_on_deactivated = s_return_on_deactivated
        
        if  type(plot_item) == pg.PlotDataItem:
            self.plot_data_item = plot_item
            self.plot_data_item.sigPlotChanged.connect(self.set_x_array_from_data_item)
            self.parent_plot_item = plot_item.parentItem()
        elif type(plot_item) == pg.PlotItem:                
            self.plot_data_item = None
            self.parent_plot_item = plot_item

        self.linear_region_item = pg.LinearRegionItem(brush = brush)
        self.linear_region_item.setZValue(ZValue)                         
        self.parent_plot_item.addItem(self.linear_region_item)
        
        self.inf_line_label = pg.InfLineLabel(self.linear_region_item.lines[label_line],
                                              self.name, position=0.78, anchor=(0.5, 0.5))
        #self.inf_line_label = pg.TextItem(self.name, anchor=(0.5, 0.5))
        self.inf_line_label.setFont(font)
        self.set_label('')
        
        if x_array == None: #give it something to work with.
            x_array = np.arange(512)
        self.set_x_array(x_array)
        
        self.start.add_listener(self.on_change_start_stop)
        self.stop.add_listener(self.on_change_start_stop)
        self.activated.add_listener(self.on_change_activated)
        self.linear_region_item.sigRegionChangeFinished.connect(self.on_change_region)

        
    
    @QtCore.Slot(object)
    def set_x_array_from_data_item(self):
        x_array,_ = self.plot_data_item.getData()
        if not np.array_equal(x_array, self.x_array):
            self.apply_new_x_array(x_array)
            #print('set_x_array_from_data_item: new x_array()')
        else:
            pass
    
    def set_x_array(self, x_array):
        '''
        use this function to update the x_array
        not required to use if type(plot_item) == pg.PlotDataItem
        '''
        self.apply_new_x_array(x_array)
        
    def apply_new_x_array(self, x_array):
        mn,mx = self.linear_region_item.getRegion()
        self.x_array = x_array        
        self.linear_region_item.setBounds( [self.x_array.min(), self.x_array.max()] )        
        self.settings['start'] = np.argmin( (x_array - mn)**2 )
        self.settings['stop'] = np.argmin( (x_array - mx)**2 )        
        
    @property
    def slice(self):
        S = self.settings
        return np.s_[ S['start'] : S['stop'] : S['step'] ]
    
    @property
    def s_(self):
        '''returns slice based on region if activated else `s_return_on_deactivated`'''
        if self.activated.val:
            return self.slice
        else:
            return self.s_return_on_deactivated
        
    @property
    def mask(self):
        Xmin = self.x_array[self.settings['start']]
        Xmax = self.x_array[self.settings['stop']]
        return (self.x_array >= Xmin) * (self.x_array <= Xmax)
    
    @property
    def m_(self):
        if self.activated.val:
            return self.mask
        else:
            s = self.s_return_on_deactivated
            Xmin = self.x_array[s.start]
            Xmax = self.x_array[s.stop]
            return (self.x_array >= Xmin) * (self.x_array <= Xmax)      
            
        
    @QtCore.Slot(object)
    def on_change_region(self):
        '''
        updates settings based on region 
        '''
        print(self.name, 'on_change_region')
        self.region_min, self.region_max = mn,mx = self.linear_region_item.getRegion()
        self.settings['start'] = np.argmin( (self.x_array - mn)**2 )
        self.settings['stop'] = np.argmin( (self.x_array - mx)**2 )
        self.region_changed_signal.emit()
        self.slicer_updated()
        
    def on_change_start_stop(self):
        '''
        updates linear_region_item based on settings 
        '''
        rgn = [0,0]
        rgn[0] = self.x_array[ self.settings['start'] ]
        try:
            rgn[1] = self.x_array[ self.settings['stop'] ]
        except IndexError:
            rgn[1] = self.x_array[ self.settings['start'] ]
        self.linear_region_item.setRegion(rgn)    
        
    def on_change_activated(self):
        activated = self.activated.val
        self.linear_region_item.setEnabled(activated)
        self.linear_region_item.setAcceptHoverEvents(activated)
        self.linear_region_item.setAcceptTouchEvents(activated)
        self.linear_region_item.setVisible(activated)
        self.region_changed_signal.emit()
        self.slicer_updated()
        
    def New_UI(self, include=None, exclude=[]):
        ui_widget = self.settings.New_UI(include, exclude)
        ui_widget.layout().insertRow(0, QtWidgets.QLabel("<b>{}</b>".format(self.name)) )  
        return ui_widget
        
        
    def set_label(self, text='', title=None, color=(200,200,200)):
        if title == None:
            title = self.name
        label = '<h3>{}</h3>{}'.format(title, text)
        self.inf_line_label.setHtml(label)
        self.inf_line_label.setColor(color)
        
    def get_sliced_xar(self, ar=None):
        if ar==None:
            if hasattr(self, 'plot_data_item'):
                _,y = self.plot_data_item.getData()
                return (self.x_array[self.s_], y[self.s_])                
            else:
                return None
        else:
            return (self.x_array[self.s_], ar[self.s_])


import logging
import subprocess
from enum import Enum

logger = logging.getLogger()


class InfoTypes(Enum):
    INFO = "INFO"
    FAILED = "FAILED"
    ERROR = "ERROR"
    SUCCESS = "SUCCESS"
    TASK = "TASK"


def decorate_info_text(text, info_type: InfoTypes):
    logger.debug(f"{info_type.value} {text}")
    if info_type == InfoTypes.INFO:
        text = f"<p style='color:black;'>INFO: {text}</p>"
    if info_type == InfoTypes.FAILED:
        text = f"<p style='color:red;'>FAILED: {text}</p>"
    if info_type == InfoTypes.ERROR:
        text = f"<p style='color:#7719aa;'>ERROR: {text}</p>"
    if info_type == InfoTypes.SUCCESS:
        text = f"<p style='color:green;'>SUCCESS: {text}</p>"
    if info_type == InfoTypes.TASK:
        text = f"<p style='color:#ac4c2c;'>TASK: {text}</p>"
    return text


def run_subprocess(cmd: str, shell=True, capture_output=True):
    # logger.debug(f'run_subprocess: {cmd} {shell=}, {capture_output=}')
    logger.debug(f'run_subprocess: {cmd}')
    cp = subprocess.run(cmd.split(), shell=shell,
                        capture_output=capture_output)
    return cp.stdout.decode(), cp.stderr.decode()



import argparse


def main():
    
    parser = argparse.ArgumentParser(description='Create a blank ScopeFoundry Microscope.')

    parser.add_argument('name', type=str,
                    help='name of microscope')
    
    parser.add_argument('--dir', required=False, help="Directory to create microscope, defaults to name")
    
    parser.add_argument('--git-init', action='store_true', 
                            required=False, help='Initialize a git repository inside microscope directory')
    

    args = parser.parse_args()
    print(args)
    print("-->", args['name'])
    
    
if __name__ == "__main__":
    main()



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''

# GOAL OF THIS FILE IS TO HANDLE LOW LEVEL COMMUNICATION
# SHOULD BE INDEPENDENT OF ANY SCOPEFOUNDRY FUNCTIONALITY.
# RUN THIS FILE TO TEST CONNECTION, TESTING COMMANDS WHILE CHECK main()




class $DEV_CLASS_NAME:

    def __init__(self, debug=False):
        self.debug = debug
        # TODO ESTABLISH CONNECTION HERE

    def write_property_x(self, value):
        raise NotImplementedError
        # USE CONNECTION TO WRITE A VALUE

    def read_property_x(self):
        raise NotImplementedError
        # USE CONNECTION TO READ A VALUE AND RETURN

    def read_data(self):
        # NOTE read_data is a too generic function name
        raise NotImplementedError
        # USE CONNECTION TO READ A VALUE AND RETURN



if __name__ == '__main__':
    print('start')
    dev = $DEV_CLASS_NAME(debug=True)
    print(dev.read_property_x())
    # print(dev.read_data())

    print('done')



import argparse
from collections import OrderedDict
from pathlib import Path

from qtpy import QtCore, QtWidgets, QtGui

from ScopeFoundry import BaseApp, LoggedQuantity, ini_io
from ScopeFoundry.helper_funcs import load_qt_ui_from_pkg, sibling_path
from .viewers.file_info import FileInfoView


class DataBrowser(BaseApp):
    
    name = "DataBrowser"
    
    def __init__(self, argv, dark_mode=False):
        BaseApp.__init__(self, argv, dark_mode)
        self.setup()
        parser = argparse.ArgumentParser()
        for lq in self.settings.as_list():
            parser.add_argument("--" + lq.name)
        args = parser.parse_args()
        for lq in self.settings.as_list():
            if lq.name in args:
                val = getattr(args, lq.name)
                if val is not None:
                    lq.update_value(val)

    def setup(self):
        
        self._setting_paths = {}

        self.views = OrderedDict()

        self.plugin_update_funcs = []
        self.plugins = {}
        self.keys = {
            QtCore.Qt.Key_Delete: self.on_recycle,
        }
        self.ctrl_keys = {
            QtCore.Qt.Key_R: self.on_rename,
        }

        s = self.settings
        s.New("data_filename", dtype="file")
        s.New("browse_dir", dtype="file", is_dir=True, initial="/")
        s.New("file_filter", dtype=str, initial="*.*,")
        s.New(
            "auto_select_view",
            dtype=bool,
            initial=True,
            description="auto selects the view when file name is changed.",
        )
        s.New("view_name", dtype=str, initial="file_info", choices=("0",))

        self._setting_paths.update(construct_lq_paths(self.settings, "app"))

        self.setup_ui()

        self.current_view = self.add_view(FileInfoView(self))
        self.setup_view(self.current_view)

        # add callbacks
        s.data_filename.add_listener(self.on_change_data_filename)
        s.data_filename.add_listener(self.on_change_data_filename_handle_plugins)

        s.browse_dir.add_listener(self.on_change_browse_dir)
        s["browse_dir"] = Path.cwd()

        s.file_filter.add_listener(self.on_change_file_filter)
        s.view_name.add_listener(self.on_change_view_name)

    def setup_ui(self):
        self.ui = load_qt_ui_from_pkg("ScopeFoundry.data_browser", "data_browser.ui")

        self.ui.setWindowTitle("ScopeFoundry: Data Browser")

        logo_path = sibling_path(__file__, "logo.png")
        self.set_logo(logo_path)

        # file system tree
        self.tree_view = self.ui.treeView
        self.fs_model = QtWidgets.QFileSystemModel()
        self.fs_model.setRootPath(QtCore.QDir.currentPath())
        self.tree_view.setModel(self.fs_model)
        self.tree_view.setIconSize(QtCore.QSize(16, 16))
        self.tree_view.setSortingEnabled(True)
        self.tree_view.setColumnWidth(0, 500)  # make name column wider
        self.tree_selectionModel = self.tree_view.selectionModel()
        self.tree_selectionModel.selectionChanged.connect(
            self.on_treeview_selection_change
        )

        # UI Connections
        s = self.settings
        s.data_filename.connect_to_browse_widgets(
            self.ui.data_filename_lineEdit, self.ui.data_filename_browse_pushButton
        )
        s.browse_dir.connect_to_browse_widgets(
            self.ui.browse_dir_lineEdit, self.ui.browse_dir_browse_pushButton
        )
        self.ui.data_filename_recycle_pushButton.clicked.connect(self.on_recycle)
        self.ui.data_filename_rename_pushButton.clicked.connect(self.on_rename)
        s.view_name.connect_bidir_to_widget(self.ui.view_name_comboBox)
        s.auto_select_view.connect_to_widget(self.ui.auto_select_checkBox)
        s.file_filter.connect_bidir_to_widget(self.ui.file_filter_lineEdit)

        self.ui.console_pushButton.clicked.connect(self.console_widget.show)
        self.ui.log_pushButton.clicked.connect(self.logging_widget.show)

        self.ui.action_save_as_ini.triggered.connect(self.settings_save_dialog)        
        self.ui.action_load_ini.triggered.connect(self.settings_load_dialog)     

        self.ui.show()
        self.ui.raise_()

        self.ui.keyPressEvent = self.handle_key_board

    def set_logo(self, logo_path):
        logo_icon = QtGui.QIcon(logo_path)
        self.qtapp.setWindowIcon(logo_icon)
        self.ui.setWindowIcon(logo_icon)

    def handle_key_board(self, event):
        if event.modifiers() != QtCore.Qt.ControlModifier:
            func = self.keys.get(event.key(), None)
            if func is not None:
                func()
            return

        func = self.ctrl_keys.get(event.key(), None)
        if func is not None:
            func()

    def add_view(self, new_view):
        self.log.debug("add_view called {}".format(new_view))

        # Update views dict
        self.views[new_view.name] = new_view
        self.settings.view_name.change_choice_list(list(self.views.keys()))

        self.log.debug("add_view done {}".format(new_view))
        print("added view", repr(new_view.name))

        return new_view

    def add_plugin(self, plugin):
        self.ui.plugin_buttons_layout.addWidget(plugin.get_show_hide_button())
        self.plugin_update_funcs.append(plugin.update_if_showing)
        self.ctrl_keys[plugin.show_keyboard_key] = plugin.toggle_show_hide
        self.plugins[plugin.name] = plugin

    def load_view(self, new_view):
        # deprecated use DataBrowser.add_view(new_view) instead
        self.add_view(new_view)

    def on_change_data_filename(self):
        fname = self.settings["data_filename"]
        if not Path(fname).is_file():
            print("invalid data_filename", fname)
            return

        print("file", fname)

        # select view
        if self.settings["auto_select_view"]:
            view_name = self.auto_select_view(fname)
            if view_name != self.current_view.name:
                self.settings["view_name"] = view_name
                self.setup_view(self.views[view_name])

        self.current_view.on_change_data_filename(fname)

    def on_change_data_filename_handle_plugins(self):
        fname = self.settings["data_filename"]
        for func in self.plugin_update_funcs:
            func(fname)

    @QtCore.Slot()
    def on_change_browse_dir(self):
        self.log.debug("on_change_browse_dir")
        self.ui.treeView.setRootIndex(self.fs_model.index(self.settings['browse_dir']))
        self.fs_model.setRootPath(self.settings['browse_dir'])

    def on_change_file_filter(self):
        self.log.debug("on_change_file_filter")
        filter_str = self.settings['file_filter']
        if filter_str == "":
            filter_str = "*"
            self.settings['file_filter'] = "*"
        filter_str_list = [x.strip() for x in filter_str.split(',')]
        self.log.debug(filter_str_list)
        self.fs_model.setNameFilters(filter_str_list)

    def setup_view(self, view):
        if view.view_loaded:
            return
        view.setup()
        self.ui.data_view_layout.addWidget(self.current_view.ui)
        view.view_loaded = True
        self._setting_paths.update(construct_lq_paths(view.settings, f"app/{view.name}"))

    def on_change_view_name(self):
        # hide previous
        self.current_view.ui.hide()

        # setup, show new
        self.current_view = self.views[self.settings["view_name"]]
        self.setup_view(self.current_view)
        self.current_view.ui.show()

        # udpate
        fname = self.settings["data_filename"]
        if Path(fname).is_file():
            self.current_view.on_change_data_filename(fname)

    def on_treeview_selection_change(self, sel, desel):
        fname = self.fs_model.filePath(self.tree_selectionModel.currentIndex())
        self.settings['data_filename'] = fname
#        print( 'on_treeview_selection_change' , fname, sel, desel)

    def auto_select_view(self, fname):
        "return the name of the last supported view for the given fname"
        for view_name, view in list(self.views.items())[::-1]:
            if view.is_file_supported(fname):
                return view_name
        return "file_info"

    def on_recycle(self):
        # import here because send2trash is not a built in library.
        import send2trash

        send2trash.send2trash(Path(self.settings["data_filename"]))

    def on_rename(self):
        fname = self.settings["data_filename"]
        dialog = RenameDialog(prev_path_name=fname)
        if not dialog.exec_():
            return  # dialog was escaped

        if dialog.new_name:
            Path(fname).rename(dialog.new_name)
            
    def read_setting(self, path, ini_string_value=False):
        lq = self.get_lq(path)
        if ini_string_value:
            return lq.ini_string_value()
        return lq.val
    
    def write_setting(self, path:str, value):
        lq = self.get_lq(path)
        if lq is not None:
            lq.update_value(value)
        else:
            print(path, "does not exist")

    def get_lq(self, path:str) -> LoggedQuantity:
        """
        returns the LoggedQuantity defined by a path string of the form 'section/[component/]setting'
        where section are "views" or "app"
        """
        parts = path.split("/")
        
        settings = self.settings
        if parts[0] in ("views"):
            view = self.views.get(parts[1], None)
            if view is not None:
                self.load_view(view)
                settings = view.settings        
        elif parts[0] == "plugins":
            plugin = self.plugins.get(parts[1], None)
            if plugin is not None:
                settings = plugin.settings
                
        if parts[-1] in settings: 
            return self.settings.get_lq(parts[1])
    
    def settings_load_ini(self, fname):
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.              
        ==============  =========  ==============================================
        """
        settings = ini_io.load_settings(fname)
        for path, value in settings.items():
            self.write_setting(path, value)
            
    def settings_save_ini(self, fname):
        settings = self.read_settings(ini_string_value=True)
        ini_io.save_settings(fname, settings)
        
    def read_settings(self, ini_string_value=False):
        """returns a dictionary (path, value) of registered settings"""
        return {p:self.read_setting(p, ini_string_value) for p in self._setting_paths}
    
    def settings_save_dialog(self):
        """Opens a save as ini dialogue in the app user interface."""
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(self.ui, "Save Settings file", "", "Settings File (*.ini)")
        if fname:
            self.settings_save_ini(fname)
    
    def settings_load_dialog(self):
        """Opens a load ini dialogue in the app user interface"""
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(self.ui, "Open Settings file", "", "Settings File (*.ini)")
        if fname.endswith(".ini"):
            self.settings_load_ini(fname)
        # elif fname.endswith(".h5"):
        #     self.settings_load_h5(fname)
        

class RenameDialog(QtWidgets.QDialog):

    def __init__(self, prev_path_name):
        QtWidgets.QDialog.__init__(self)

        self.new_name = None

        rename_btn = QtWidgets.QPushButton("rename")
        cancel_btn = QtWidgets.QPushButton("cancel")
        self.new_name_w = QtWidgets.QLineEdit(prev_path_name)

        cancel_btn.clicked.connect(self.on_cancel)
        rename_btn.clicked.connect(self.on_rename)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.new_name_w)
        layout.addWidget(rename_btn)
        layout.addWidget(cancel_btn)
        self.setLayout(layout)

        self.setWindowTitle("rename")
        self.setMinimumWidth(500)

    def on_cancel(self):
        self.new_name = None
        self.accept()

    def on_rename(self):
        self.new_name = self.new_name_w.text()
        self.accept()

        
def construct_lq_paths(lq_collection, pre="app"):
    new = {}
    for name, lq in lq_collection.as_dict().items():
        path = f"{pre}/{name}"
        lq.set_path(path)
        new[path] = lq
    return new



'''
Created on Feb 3, 2023

@author: Benedikt Ursprung
'''
import logging

from ScopeFoundry.base_app import BaseMicroscopeApp

logging.basicConfig(level=logging.DEBUG)

class App(BaseMicroscopeApp):

    name = 'test_app'

    def setup(self):

        from ScopeFoundry.controlling import PIDFeedbackControl
        self.add_measurement(PIDFeedbackControl(self))
        from ScopeFoundry.controlling import RangedOptimization
        self.add_measurement(RangedOptimization(self))
        
    def setup_ui(self):
        from qtpy.QtWidgets import QWidget, QVBoxLayout
        widget = QWidget()
        widget.setMaximumWidth(380)
        layout = QVBoxLayout(widget)
        self.add_quickbar(widget)
        layout.addWidget(self.measurements['pid_feedback_control'].New_mini_UI())

if __name__ == '__main__':
    import sys
    app = App(sys.argv)
    sys.exit(app.exec_())



from typing import Union
from typing_extensions import Self

from qtpy.QtWidgets import QListWidgetItem

from ScopeFoundry.measurement import Measurement

VisitReturnType = Union[Self, None]  # return either a go-to Item or None if next


class BaseItem(QListWidgetItem):
    item_type = "item - overwrite me"

    def __init__(self, measure:Measurement, **kwargs):
        super().__init__()
        self.app = measure.app
        self.measure = measure
        self.kwargs = kwargs
        self._update_appearance()

    def visit(self) -> VisitReturnType:
        raise NotImplementedError

    def _update_appearance(self, text=None) -> str:
        if text == None:
            kwargs_str = " ".join([f"{val}" for val in self.kwargs.values()])
            text = f"{self.item_type}: {kwargs_str}"
        self.setText(text)
        return text

    def reset(self) -> None:
        pass



from jinja2 import Environment, FileSystemLoader

def load_template_by_dir(template_dir, filename, context):
    TEMPLATE_ENVIRONMENT = Environment(
    autoescape=False,
    loader=FileSystemLoader(template_dir),
    trim_blocks=False)
    return TEMPLATE_ENVIRONMENT.get_template(filename).render(context)



from ScopeFoundry import BaseMicroscopeApp, HardwareComponent
import logging

class TestFailHW(HardwareComponent):
    
    name = 'fail_hw'

    def setup(self):
        print(self.name, "setup")

        self.settings.New('fail_on_connect', dtype=bool)
        self.settings.New('fail_on_disconnect', dtype=bool, initial=False)

    def connect(self):
        print(self.name, "connect")  
        if self.settings['fail_on_connect']:     
            raise IOError("CONNECT FAIL!")

    def disconnect(self):
        print(self.name, "disconnect")        
        if self.settings['fail_on_disconnect']:     
            raise IOError("DISCONNECT FAIL!")


class TestApp(BaseMicroscopeApp):
    
    name = 'test_app'
    
    def setup(self):
        
        self.add_hardware(TestFailHW(self))
        logging.getLogger("LoggedQuantity").setLevel('DEBUG')
    
    
if __name__ == '__main__':
    import sys
    app = TestApp(sys.argv)
    
    app.exec_()
    


from ScopeFoundry.data_browser import DataBrowserView
from ScopeFoundry import LQCollection
import pyqtgraph as pg
import pyqtgraph.dockarea as dockarea
import numpy as np
from qtpy import QtWidgets, QtGui, QtCore
from ScopeFoundry.widgets import RegionSlicer
import h5py
import os
import time
import datetime
from ScopeFoundry.helper_funcs import sibling_path

class HyperSpectralBaseView(DataBrowserView):
    
    name = 'HyperSpectralBaseView'
    
    def setup(self):

        ## Dummy data Structures (override in func:self.load_data())
        self.hyperspec_data = np.arange(10*10*34).reshape( (10,10,34) )
        self.display_image = self.hyperspec_data.sum(-1)# np.random.rand(10,10)
        self.spec_x_array = np.arange(34)

        # Call :func:set_scalebar_params() during self.load_data() to add a scalebar!
        self.scalebar_type = None


        # Will be filled derived maps and images
        self.display_images = dict()
        self.spec_x_arrays = dict()   

        ## Graphs and Interface 
        self.ui = self.dockarea = dockarea.DockArea()
        self.imview = pg.ImageView()
        self.imview.getView().invertY(False) # lower left origin
        self.image_dock = self.dockarea.addDock(name='Image', widget=self.imview)
        self.graph_layout = pg.GraphicsLayoutWidget()
        self.spec_dock = self.dockarea.addDock(name='Spec Plot', widget=self.graph_layout)

        self.line_colors = ['w', 'r', 'g', 'b', 'y', 'm', 'c']
        self.spec_plot = self.graph_layout.addPlot()
        self.spec_plot.setLabel('left', 'Intensity', units='counts') 
        self.rect_plotdata = self.spec_plot.plot(y=[0,2,1,3,2], pen=self.line_colors[0])
        self.point_plotdata = self.spec_plot.plot(y=[0,2,1,3,2], pen=self.line_colors[1])
        self.point_plotdata.setZValue(-1)

        #correlation plot
        self.corr_layout = pg.GraphicsLayoutWidget()
        self.corr_plot = self.corr_layout.addPlot()
        self.corr_plotdata = pg.ScatterPlotItem(x=[0,1,2,3,4], y=[0,2,1,3,2], size=17, 
                                        pen=pg.mkPen(None), brush=pg.mkBrush(255, 255, 255, 60))
        self.corr_plot.addItem(self.corr_plotdata)        
        self.corr_dock = self.dockarea.addDock(name='correlation', widget=self.corr_layout, 
                              position='below',  relativeTo = self.spec_dock)
        self.corr_plotdata.sigClicked.connect(self.corr_plot_clicked)
        self.spec_dock.raiseDock()
        
        # Rectangle ROI
        self.rect_roi = pg.RectROI([20, 20], [20, 20], pen=self.line_colors[0])
        self.rect_roi.addTranslateHandle((0.5,0.5))        
        self.imview.getView().addItem(self.rect_roi)        
        self.rect_roi.sigRegionChanged[object].connect(self.on_change_rect_roi)
        
        # Point ROI
        self.circ_roi = pg.CircleROI( (0,0), (2,2) , movable=True, pen=self.line_colors[1])
        #self.circ_roi.removeHandle(self.circ_roi.getHandles()[0])
        h = self.circ_roi.addTranslateHandle((0.5,.5))
        h.pen = pg.mkPen(pen=self.line_colors[1])
        h.update()
        self.imview.getView().addItem(self.circ_roi)
        self.circ_roi.removeHandle(0)
        self.circ_roi_plotline = pg.PlotCurveItem([0], pen=self.line_colors[1])
        self.imview.getView().addItem(self.circ_roi_plotline) 
        self.circ_roi.sigRegionChanged[object].connect(self.on_update_circ_roi)
                
        #font
        font = QtGui.QFont("Times", 12)
        font.setBold(True)
            
        #settings
        self.default_display_image_choices = ['default', 'sum']
        self.settings.New('display_image', str, choices = self.default_display_image_choices, initial = 'default')    
        self.settings.display_image.add_listener(self.on_change_display_image)    
        
        self.default_x_axis_choices = ['default', 'index']
        self.x_axis = self.settings.New('x_axis', str, initial = 'default', choices = self.default_x_axis_choices)
        self.x_axis.add_listener(self.on_change_x_axis)

        self.norm_data = self.settings.New('norm_data', bool, initial = False)
        self.bg_subtract = self.settings.New('bg_subtract', str, initial='None', choices=('None', 'bg_slice', 'costum_const'))
        self.bg_counts = self.settings.New('bg_value', initial=0, unit='cts/bin')
        
        self.settings.New('default_view_on_load', bool, initial=True)
        
        self.binning = self.settings.New('binning', int, initial = 1, vmin=1)

        self.spatial_binning = self.settings.New('spatial_binning', int, initial = 1, vmin=1)

        self.cor_X_data = self.settings.New('cor_X_data', str, choices = self.default_display_image_choices,
                                            initial = 'default')
        self.cor_Y_data = self.settings.New('cor_Y_data', str, choices = self.default_display_image_choices,
                                            initial = 'sum')


        # data slicers
        self.x_slicer = RegionSlicer(self.spec_plot, name='x_slice', 
                                      brush = QtGui.QColor(0,255,0,70), 
                                      ZValue=10, font=font, initial=[100,511])
        self.bg_slicer = RegionSlicer(self.spec_plot, name='bg_slice', slicer_updated_func=self.update_display,
                                      brush = QtGui.QColor(255,255,255,70), 
                                      ZValue=11, font=font, initial=[0,80], label_line=0)
        
        
        # peakutils
        self.peakutils_settings = LQCollection()    
        self.show_peak_line = self.peakutils_settings.New('show_peak_line', bool, initial=False)
        self.show_peak_line.add_listener(self.update_display)
        self.baseline_deg = self.peakutils_settings.New('baseline_deg', int, initial=0, vmin=-1, vmax=100)
        self.baseline_deg.add_listener(self.update_display)
        self.thres = self.peakutils_settings.New('thres', float, initial=0.5, vmin=0, vmax=1) 
        self.thres.add_listener(self.update_display)
        self.peakutils_settings.New('unique_solution', bool, initial=False)
        self.peakutils_settings.New('min_dist', int, initial=-1)
        self.peakutils_settings.New('gaus_fit_refinement', bool, initial=True)
        self.peakutils_settings.New('ignore_phony_refinements', bool, initial=True)
        self.base_plotdata = self.spec_plot.plot(y=[0,2,1,3,2], pen=self.line_colors[2])
        self.peak_lines = []


        # Settings Docks
        self.settings_dock = self.dockarea.addDock(name='settings', position='left', relativeTo=self.image_dock)
        self.settings_widgets = [] # Hack part 1/2: allows to use settings.New_UI() and have settings defined in scan_specific_setup()
        self.settings_widgets.append(self.x_slicer.New_UI())
        self.settings_widgets.append(self.bg_slicer.New_UI())
        self.scan_specific_setup()
        self.generate_settings_ui() # Hack part 2/2: Need to generate settings after scan_specific_setup()
            
        self.peakutils_dock = self.dockarea.addDock(name='PeakUtils', position='below', relativeTo=self.settings_dock)
        self.peakutils_dock.addWidget(self.peakutils_settings.New_UI())
        self.settings_dock.raiseDock()
             
        self.settings_dock.setStretch(x=0,y=0)
        self.peakutils_dock.setStretch(x=0,y=0)
        self.image_dock.setStretch(x=100,y=1)

        #self.settings_dock.widgetArea.setStyleSheet('Dock > QWidget {border:0px; border-radius:0px}')
        #self.peakutils_dock.widgetArea.setStyleSheet('Dock > QWidget {border:0px; border-radius:0px}')
        
        ### LQ Connections
        self.norm_data.add_listener(self.update_display)
        self.bg_subtract.add_listener(self.update_display)
        self.bg_counts.add_listener(self.update_display)
        self.binning.add_listener(self.update_display)
        self.spatial_binning.add_listener(self.bin_spatially)
        self.cor_X_data.add_listener(self.on_change_corr_settings)
        self.cor_Y_data.add_listener(self.on_change_corr_settings)

        self.bg_slicer.activated.add_listener(lambda:self.bg_subtract.update_value('bg_slice') if self.bg_slicer.activated.val else None)        

        self.show_lines = ['show_circ_line','show_rect_line']
        for x in self.show_lines:
            lq = self.settings.New(x, bool, initial=True)
            lq.add_listener(self.on_change_show_lines)        



    
    def generate_settings_ui(self):
        self.settings_ui = self.settings.New_UI()
        self.settings_dock.addWidget(self.settings_ui)        

        #some more self.settings_widgets[]
        #self.update_display_pushButton = QtWidgets.QPushButton(text = 'update display')
        #self.settings_widgets.append(self.update_display_pushButton)
        #self.update_display_pushButton.clicked.connect(self.update_display)  

        self.default_view_pushButton = QtWidgets.QPushButton(text = 'default img view')
        self.settings_widgets.append(self.default_view_pushButton)
        self.default_view_pushButton.clicked.connect(self.default_image_view) 
        
        self.recalc_median_pushButton = QtWidgets.QPushButton(text = 'recalc median map')
        self.settings_widgets.append(self.recalc_median_pushButton)
        self.recalc_median_pushButton.clicked.connect(self.recalc_median_map)

        self.recalc_sum_pushButton = QtWidgets.QPushButton(text = 'recalc sum map')
        self.settings_widgets.append(self.recalc_sum_pushButton)
        self.recalc_sum_pushButton.clicked.connect(self.recalc_sum_map)
        
        self.recalc_peak_map_pushButton = QtWidgets.QPushButton(text = 'recalc PeakUtils map')
        self.settings_widgets.append(self.recalc_peak_map_pushButton)
        self.recalc_peak_map_pushButton.clicked.connect(self.recalc_peak_map)

        self.save_state_pushButton = QtWidgets.QPushButton(text = 'save_state')
        self.settings_widgets.append(self.save_state_pushButton)
        self.save_state_pushButton.clicked.connect(self.save_state)

        self.delete_current_display_image_pushButton = QtWidgets.QPushButton(text = 'remove image')
        self.settings_widgets.append(self.delete_current_display_image_pushButton)
        self.delete_current_display_image_pushButton.clicked.connect(self.delete_current_display_image)
        
        # Place the self.settings_widgets[] on a grid
        ui_widget =  QtWidgets.QWidget()
        gridLayout = QtWidgets.QGridLayout()
        gridLayout.setSpacing(0)
        gridLayout.setContentsMargins(0, 0, 0, 0)
        ui_widget.setLayout(gridLayout)        
        for i,w in enumerate(self.settings_widgets):
            gridLayout.addWidget(w, int(i/2), i%2)
        self.settings_dock.addWidget(ui_widget)          
        self.settings_dock.layout.setSpacing(0)                     
        
    def add_spec_x_array(self, key, array):
        self.spec_x_arrays[key] = array
        self.settings.x_axis.add_choices(key, allow_duplicates=False)

    def add_display_image(self, key, image):
        key = self.add_descriptor_suffixes(key)
        self.display_images[key] = image
        self.settings.display_image.add_choices(key, allow_duplicates=False)
        self.cor_X_data.change_choice_list(self.display_images.keys())
        self.cor_Y_data.change_choice_list(self.display_images.keys())
        self.cor_X_data.update_value(self.cor_Y_data.val)
        self.cor_Y_data.update_value(key)
        self.on_change_corr_settings()
    
    def add_descriptor_suffixes(self, key):
        if self.x_slicer.activated.val:
            key += '_x{}-{}'.format(self.x_slicer.start.val, self.x_slicer.stop.val)
        if self.settings['bg_subtract'] == 'bg_slice' and self.bg_slicer.activated.val:
            key += '_bg{}-{}'.format(self.bg_slicer.start.val, self.bg_slicer.stop.val)
        if self.settings['bg_subtract'] == 'costum_count':
            key += '_bg{1.2f}'.format(self.bg_counts.val)
        return key        
    
    def delete_current_display_image(self):
        key = self.settings.display_image.val
        del self.display_images[key]
        self.settings.display_image.remove_choices(key)
        self.cor_X_data.remove_choices(key)
        self.cor_Y_data.remove_choices(key)

            
    def get_xy(self, ji_slice, apply_use_x_slice=False):
        '''
        returns processed hyperspec_data averaged over a given spatial slice.
        '''
        x,hyperspec_dat = self.get_xhyperspec_data(apply_use_x_slice)
        y = hyperspec_dat[ji_slice].mean(axis=(0,1))
        #self.databrowser.ui.statusbar.showMessage('get_xy(), counts in slice: {}'.format( y.sum() ) )

        if self.settings['norm_data']:
            y = norm(y)          
        return (x,y)
    
    def get_bg(self):
        bg_subtract_mode = self.bg_subtract.val
        if bg_subtract_mode == 'bg_slice':
            if not self.bg_slicer.activated:
                self.bg_slicer.activated.update_value(True)
            bg_slice = self.bg_slicer.slice
            bg = self.hyperspec_data[:,:,bg_slice].mean()
            self.bg_slicer.set_label(title=bg_subtract_mode,
                text='{:1.1f} cts<br>{} bins'.format(bg,bg_slice.stop-bg_slice.start))
        elif bg_subtract_mode == 'costum_const':
            bg = self.bg_counts.val
            self.bg_slicer.set_label('', title=bg_subtract_mode)
        else:
            bg = 0
            self.bg_slicer.set_label('', title=bg_subtract_mode)
        return bg
        
    def get_xhyperspec_data(self, apply_use_x_slice=True):
        '''
        returns processed hyperspec_data
        '''
        bg = self.get_bg()
        hyperspec_data = self.hyperspec_data
        x = self.spec_x_array
        if apply_use_x_slice and self.x_slicer.activated.val:
            x = x[self.x_slicer.slice]
            hyperspec_data = hyperspec_data[:,:,self.x_slicer.slice]
        binning = self.settings['binning']
        if  binning!= 1:
            x,hyperspec_data = bin_y_average_x(x, hyperspec_data, binning, -1, datapoints_lost_warning=False)
            bg *= binning
        msg = 'effective subtracted bg value is binnging*bg ={:0.1f} which is up to {:2.1f}% of max value.'.format(bg, bg/np.max(hyperspec_data)*100 )
        self.databrowser.ui.statusbar.showMessage(msg)
        if self.settings['norm_data']:
            return (x,norm_map(hyperspec_data-bg))
        else:
            return (x,hyperspec_data-bg)
    
    def on_change_x_axis(self):
        key = self.settings['x_axis']
        if key in self.spec_x_arrays:
            self.spec_x_array = self.spec_x_arrays[key]
            self.x_slicer.set_x_array(self.spec_x_array)
            self.bg_slicer.set_x_array(self.spec_x_array)            
            self.spec_plot.setLabel('bottom', key)
            self.update_display()
            
    def on_change_display_image(self):
        key = self.settings['display_image']
        if key in self.display_images:
            self.display_image = self.display_images[key]
            self.update_display()
        if self.display_image.shape == (1,1):
            self.databrowser.ui.statusbar.showMessage('Can not display single pixel image!')
                
    def scan_specific_setup(self):
        #override this!
        pass
        
    def is_file_supported(self, fname):
        # override this!
        return False

    def post_load(self):
        # override this!
        pass    
    
    def on_change_data_filename(self, fname):
        if fname == "0":
            return
        self.reset()
        try:
            self.scalebar_type = None
            self.load_data(fname)
            if self.settings['spatial_binning'] != 1:
                self.hyperspec_data = bin_2D(self.hyperspec_data, self.settings['spatial_binning'])
                self.display_image = bin_2D(self.display_image, self.settings['spatial_binning'])
            self.display_images['default'] = self.display_image
            self.display_images['sum'] = self.hyperspec_data.sum(axis=-1)         
            self.spec_x_arrays['default'] = self.spec_x_array
            self.spec_x_arrays['index'] = np.arange(self.hyperspec_data.shape[-1])
            self.databrowser.ui.statusbar.clearMessage()
            self.post_load()
            self.add_scalebar()
        except Exception as err:
            HyperSpectralBaseView.load_data(self, fname) # load default dummy data
            self.databrowser.ui.statusbar.showMessage("failed to load {}: {}".format(fname, err))
            raise(err)
        finally:
            self.on_change_display_image()
            self.on_change_corr_settings()
            self.update_display()
        self.on_change_x_axis()

        if self.settings['default_view_on_load']:
            self.default_image_view()   
            
    def add_scalebar(self):
        ''' not intended to use: Call set_scalebar_params() during load_data()'''
        if hasattr(self, 'scalebar'):
            self.imview.getView().removeItem(self.scalebar)
            del self.scalebar
        if self.scalebar_type == 'ConfocalScaleBar':
            from viewers.scalebars import ConfocalScaleBar
            num_px = self.display_image.shape[1] #horizontal dimension!
            kwargs = self.scalebar_kwargs       
            self.scalebar = ConfocalScaleBar(num_px=num_px, 
                                **kwargs)
            self.scalebar.setParentItem(self.imview.getView())
            self.scalebar.anchor((1, 1), (1, 1), offset=kwargs['offset'])
        elif self.scalebar_type == None:
            self.scalebar = None
            
    def set_scalebar_params(self, h_span, units='m', scalebar_type='ConfocalScaleBar',
                           stroke_width=10, brush='w', pen='k', offset=(-20, -20)):
        '''
        call this function during load_data() to add a scalebar!
        *h_span*  horizontal length of image in units of *units* if positive.
                  Else, scalebar is in units of pixels (*units* ignored).
        *units*   SI length unit of *h_span*.
        *scalebar_type* is either `None` (no scalebar will be added)
          or `"ConfocalScaleBar"` (default).
        *stroke_width*, *brush*, *pen* and *offset* affect appearance and 
         positioning of the scalebar.
        '''
        assert scalebar_type in [None, 'ConfocalScaleBar']
        self.scalebar_type = scalebar_type
        span_meter = {'m':1, 'cm':1e-2, 'mm':1e-3, 'um':1e-6, 
                      'nm':1e-9, 'pm':1e-12, 'fm':1e-15}[units] * h_span
        self.scalebar_kwargs = {'span':span_meter, 'brush':brush, 'pen':pen,
                                'width':stroke_width, 'offset':offset}

    def update_display(self):
        # pyqtgraph axes are (x,y), but display_images are in (y,x) so we need to transpose        
        if self.display_image is not None:
            self.imview.setImage(self.display_image.T)
            self.on_change_rect_roi()
            self.on_update_circ_roi()
            
    def reset(self):
        '''
        resets the dictionaries
        '''
        keys_to_delete = list( set(self.display_images.keys()) - set(self.default_display_image_choices) )
        for key in keys_to_delete:
            del self.display_images[key]
        keys_to_delete = list( set(self.spec_x_arrays.keys()) - set(self.default_x_axis_choices) )
        for key in keys_to_delete:
            del self.spec_x_arrays[key]
        self.settings.display_image.change_choice_list(self.default_display_image_choices)
        self.settings.x_axis.change_choice_list(self.default_x_axis_choices)

    
    def load_data(self, fname):
        """
        override to set hyperspectral dataset and the display image
        need to define:
            * self.hyperspec_data (shape Ny, Nx, Nspec)
            * self.display_image (shape Ny, Nx)
            * self.spec_x_array (shape Nspec)
        """
        self.hyperspec_data = np.arange(10*10*34).reshape( (10,10,34) )
        self.display_image = self.hyperspec_data.sum(-1)
        self.spec_x_array = np.arange(34)
    
    @QtCore.Slot(object)
    def on_change_rect_roi(self, roi=None):
        # pyqtgraph axes are (x,y), but hyperspec is in (y,x,spec) hence axes=(1,0)      
        roi_slice, roi_tr = self.rect_roi.getArraySlice(self.hyperspec_data, self.imview.getImageItem(), axes=(1,0)) 
        self.rect_roi_slice = roi_slice
        x,y = self.get_xy(roi_slice, apply_use_x_slice=False)
        self.rect_plotdata.setData(x, y)
        self.on_change_corr_settings()
        self.update_peaks(*self.get_xy(roi_slice, apply_use_x_slice=True), pen=self.line_colors[0])

        
    @QtCore.Slot(object)        
    def on_update_circ_roi(self, roi=None):
        if roi is None:
            roi = self.circ_roi

        roi_state = roi.saveState()
        x0, y0 = roi_state['pos']
        xc = x0 + 1
        yc = y0 + 1

        Ny, Nx, Nspec = self.hyperspec_data.shape
        
        i = max(0, min(int(xc),  Nx-1))
        j = max(0, min(int(yc),  Ny-1))
                
        self.circ_roi_plotline.setData([xc, i+0.5], [yc, j + 0.5])
        
        self.circ_roi_ji = (j,i)    
        self.circ_roi_slice = np.s_[j:j+1,i:i+1]
        x,y = self.get_xy(self.circ_roi_slice, apply_use_x_slice=False)  
        self.point_plotdata.setData(x, y)
        self.on_change_corr_settings()
        self.update_peaks(*self.get_xy(self.circ_roi_slice, apply_use_x_slice=True), pen=self.line_colors[1])
        
        
    def update_peaks(self, wls, spec, pen='g'):
        self.base_plotdata.setVisible( self.peakutils_settings['show_peak_line'] )
        
        for l in self.peak_lines:
            self.spec_plot.removeItem(l)
            l.deleteLater()
        self.peak_lines = []
                    
        if self.peakutils_settings['show_peak_line']:
            PS = self.peakutils_settings
            import peakutils
            base = 1.0*peakutils.baseline(spec, PS['baseline_deg'])
            self.base_plotdata.setData(wls, base)
            self.base_plotdata.setPen(pen)
            if PS['min_dist'] < 0: 
                min_dist = int(len(wls)/2) 
            else:
                min_dist = PS['min_dist']
            peaks_ = peaks(spec-base, wls, PS['thres'], PS['unique_solution'], min_dist, 
                           PS['gaus_fit_refinement'], PS['ignore_phony_refinements'])
            for p in np.atleast_1d(peaks_):
                l = pg.InfiniteLine(pos=(p,0),
                                    movable=False, angle=90, pen=pen, label='PeakUtils {value:0.2f}', 
                         labelOpts={'color':pen, 'movable': True, 'fill': (200, 200, 200, 100)})
                self.spec_plot.addItem(l)
                self.peak_lines.append(l)

    def on_change_show_lines(self):
        self.point_plotdata.setVisible(self.settings['show_circ_line'])
        self.rect_plotdata.setVisible(self.settings['show_rect_line'])
 
        
    def default_image_view(self):
        'sets rect_roi congruent to imageItem and optimizes size of imageItem to fit the ViewBox'
        iI = self.imview.imageItem
        h,w  = iI.height(), iI.width()       
        self.rect_roi.setSize((w,h))
        self.rect_roi.setPos((0,0))
        self.imview.getView().enableAutoRange()
        self.spec_plot.enableAutoRange()
        
    def recalc_median_map(self):
        x,hyperspec_data = self.get_xhyperspec_data(apply_use_x_slice=True)
        median_map = spectral_median_map(hyperspec_data,x)
        self.add_display_image('median_map', median_map)
        
    def recalc_sum_map(self):
        x,hyperspec_data = self.get_xhyperspec_data(apply_use_x_slice=True)
        _sum = hyperspec_data.sum(-1)
        self.add_display_image('sum', _sum)
        
    def recalc_peak_map(self):
        x,hyperspec_data = self.get_xhyperspec_data(apply_use_x_slice=True)
        PS = self.peakutils_settings
        _map = peak_map(hyperspec_data, x, PS['thres'], int(len(x)/2), 
                        PS['gaus_fit_refinement'], PS['ignore_phony_refinements'])
        map_name = 'peak_map'
        if  PS['gaus_fit_refinement']: 
            map_name += '_refined'
            if PS['ignore_phony_refinements']:
                map_name += '_ignored'
        self.add_display_image(map_name, _map)
          
    def on_change_corr_settings(self):
        try:
            xname = self.settings['cor_X_data']
            yname = self.settings['cor_Y_data']
            X = self.display_images[xname]
            Y = self.display_images[yname]

            #Note, the correlation plot is a dimensionality reduction 
            # (i,j,X,Y) --> (X,Y). To map the scatter points back to the image
            # we need to associate every (X,Y) on the correlation plot with 
            # their indices (i,j); in particular 
            # indices = [(j0,i0), (j0,i1), ...]
            indices = list( np.indices((X.shape)).reshape(2,-1).T )
            self.corr_plotdata.setData(X.flat, Y.flat, brush=pg.mkBrush(255, 255, 255, 50),
                                       pen=None, data=indices)

            # mark points within rect_roi 
            mask = np.zeros_like(X, dtype=bool)
            mask[self.rect_roi_slice[0:2]] = True
            cor_x = X[mask].flatten()
            cor_y = Y[mask].flatten()
            self.corr_plotdata.addPoints(cor_x, cor_y, brush=pg.mkBrush(255, 255, 204, 60), 
                    pen=pg.mkPen(self.line_colors[0], width=0.5))
            
            # mark circ_roi point
            j,i = self.circ_roi_ji
            x_circ, y_circ = np.atleast_1d(X[j,i]), np.atleast_1d(Y[j,i])
            self.corr_plotdata.addPoints(x=x_circ, y=y_circ, 
                    pen=pg.mkPen(self.line_colors[1], width=3))

            ##some more plot details 
            #self.corr_plot.getViewBox().setRange(xRange=(cor_x.min(), cor_x.max()),
            #                                     yRange=(cor_y.min(), cor_y.max()))
            
            self.corr_plot.autoRange()
            self.corr_plot.setLabels(**{'bottom':xname,'left':yname})
            from scipy.stats import spearmanr

            sm = spearmanr(cor_x, cor_y)
            text = 'Pearson\'s corr: {:.3f}<br>Spearman\'s: corr={:.3f}, pvalue={:.3f}'.format(
                                np.corrcoef(cor_x,cor_y)[0,1], sm.correlation, sm.pvalue)
            self.corr_plot.setTitle(text)
            
        except Exception as err:
            self.databrowser.ui.statusbar.showMessage('Error in on_change_corr_settings: {}'.format(err))

    def corr_plot_clicked(self, plotitem, points):
        '''
        call back function to locate a point on the correlation plot on the image. 
        
        *points* is a list of <pg.ScatterPlotItem.SpotItem> under the mouse 
        pointer during click event. For points within the rect_roi, there are 
        two items representing a pixel, but only most button one contains the
        (j,i) as data.    
        '''
        j,i = points[-1].data()         
        self.circ_roi.setPos(i-0.5,j-0.5)
                
        
    def bin_spatially(self):
        if not (self.settings['display_image'] in self.default_display_image_choices):
            self.settings.display_image.update_value( self.default_display_image_choices[0] )
        fname = self.databrowser.settings['data_filename']
        self.on_change_data_filename(fname)
    
    def save_state(self):
        from ScopeFoundry import h5_io
        fname = self.databrowser.settings['data_filename']
        view_state_fname = '{fname}_state_view_{timestamp:%y%m%d_%H%M%S}.{ext}'.format(
            fname = fname.strip('.h5'),
            timestamp=datetime.datetime.fromtimestamp(time.time()),
            ext='h5')
        h5_file = h5py.File(name = view_state_fname)
        
        with h5_file as h5_file:
            h5_group_display_images = h5_file.create_group('display_images')
            for k,v in self.display_images.items():
                h5_group_display_images.create_dataset(k, data=v)
            h5_group_spec_x_array = h5_file.create_group('spec_x_arrays')
            for k,v in self.spec_x_arrays.items():
                h5_group_spec_x_array.create_dataset(k, data=v)
            h5_group_settings_group = h5_file.create_group('settings')
            h5_io.h5_save_lqcoll_to_attrs(self.settings, h5_group_settings_group)
            h5_group_settings_group = h5_file.create_group('x_slicer_settings')
            h5_io.h5_save_lqcoll_to_attrs(self.x_slicer.settings, h5_group_settings_group)            
            h5_group_settings_group = h5_file.create_group('bg_slicer_settings')
            h5_io.h5_save_lqcoll_to_attrs(self.bg_slicer.settings, h5_group_settings_group)
            self.view_specific_save_state_func(h5_file)
            h5_file.close()          

    def view_specific_save_state_func(self, h5_file):
        '''
        you can override me, use 'h5_file' - it's already open 
        e.g:  h5_file.create_group('scan_specific_settings')
         ...
        '''
        pass
    
    def load_state(self, fname_idx=-1):
        
        # does not work properly, maybe because the order the settings are set matters?
        path = sibling_path(self.databrowser.settings['data_filename'], '')
        pre_state_fname = self.databrowser.settings['data_filename'].strip(path).strip('.h5')
        
        state_files = []
        for x in os.listdir(path):
            if pre_state_fname in x:
                if 'state_view' in x:
                    state_files.append(x)
            
        print('state_files', state_files)
        
        if len(state_files) != 0:
            h5_file = h5py.File(path + state_files[fname_idx])
            for k,v in h5_file['bg_slicer_settings'].attrs.items():
                try:
                    self.bg_slicer.settings[k] = v
                except:
                    pass    
            for k,v in h5_file['x_slicer_settings'].attrs.items():
                try:
                    self.x_slicer.settings[k] = v
                except:
                    pass
                
            for k,v in h5_file['settings'].attrs.items():
                try:
                    self.settings[k] = v
                except:
                    pass

            for k,v in h5_file['biexponential_settings'].attrs.items():
                self.biexponential_settings[k] = v
            for k,v in h5_file['export_settings'].attrs.items():
                self.export_settings[k] = v
            
            h5_file.close()
            print('loaded', state_files[fname_idx])
        
def spectral_median(spec, wls, count_min=200):
    int_spec = np.cumsum(spec)
    total_sum = int_spec[-1]
    if total_sum > count_min:
        pos = int_spec.searchsorted( 0.5*total_sum)
        wl = wls[pos]
    else:
        wl = 0
    return wl
def spectral_median_map(hyperspectral_data, wls):
    return np.apply_along_axis(spectral_median, -1, hyperspectral_data, wls=wls)

def norm(x):
    x_max = x.max()
    if x_max==0:
        return x*0.0
    else:
        return x*1.0/x_max
def norm_map(map_):
    return np.apply_along_axis(norm, -1, map_)

def bin_y_average_x(x, y, binning = 2, axis = -1, datapoints_lost_warning = True):
    '''
    y can be a n-dim array with length on axis `axis` equal to len(x)
    '''    
    new_len = int(x.__len__()/binning) * binning
    
    data_loss = x.__len__() - new_len
    if data_loss != 0 and datapoints_lost_warning:
        print('bin_y_average_x() warining: lost final', data_loss, 'datapoints')
    
    def bin_1Darray(arr, binning=binning, new_len=new_len):
        return arr[:new_len].reshape((-1,binning)).sum(1)
    
    x_ = bin_1Darray(x) / binning
    y_ = np.apply_along_axis(bin_1Darray,axis,y)
    
    return x_, y_

def bin_2D(arr,binning=2):
    '''
    bins an array of at least 2 dimension along the axis 0 and 1
    '''
    shape = arr.shape
    new_dim = int(shape[0]/binning)
    salvaged_along_dim = new_dim*binning
    lost_lines_0 = shape[0]-salvaged_along_dim
    arr = arr[0:salvaged_along_dim].reshape((-1,binning,shape[1],*shape[2:])).sum(1)
    shape = arr.shape
    new_dim = int(shape[1]/binning)
    salvaged_along_dim = new_dim*binning
    lost_lines_1 = shape[1]-salvaged_along_dim
    arr = arr[:,0:salvaged_along_dim].reshape((shape[0],-1,binning,*shape[2:])).sum(2)
    if (lost_lines_1 + lost_lines_0) >0 :
        print('cropped data:', (lost_lines_0,lost_lines_1), 'lines lost' )
    return arr

def peaks(spec, wls, thres=0.5, unique_solution=True, 
          min_dist=-1, refinement=True, ignore_phony_refinements=True):
    import peakutils
    indexes = peakutils.indexes(spec, thres, min_dist=min_dist)
    if unique_solution:
        #we only want the highest amplitude peak here!
        indexes = [indexes[spec[indexes].argmax()]]
        
    if refinement:
        peaks_x = peakutils.interpolate(wls, spec, ind=indexes)
        if ignore_phony_refinements:
            for i,p in enumerate(peaks_x):
                if p < wls.min() or p > wls.max():
                    print('peakutils.interpolate() yielded result outside wls range, returning unrefined result')
                    peaks_x[i] = wls[indexes[i]]
    else:
        peaks_x = wls[indexes]

    if unique_solution:
        return peaks_x[0]
    else:
        return peaks_x

def peak_map(hyperspectral_data, wls, thres, min_dist, refinement, ignore_phony_refinements):
    return np.apply_along_axis(peaks, -1, hyperspectral_data, 
                               wls=wls, thres=thres, 
                               unique_solution=True,
                               min_dist=min_dist, refinement=refinement,
                               ignore_phony_refinements=ignore_phony_refinements)


from .new_hw import ComTypes, new_hw
from .new_hw_qt_view import NewHWView as View
from .utils import InfoTypes, decorate_info_text


class NewHWController:
    def __init__(self, view: View):
        self.view = view
        view.set_com_types_options([t.value for t in ComTypes])
        view.model_text_changed(self.non_overwrite_guess)
        view.pretty_name_text_changed(self.on_pretty_name_set)
        view.guess_button_clicked(self.guess_pretty_name)
        view.com_texted_changed(self.on_change_com_type)
        view.create_clicked(self.on_create)
        view.set_overwrite(False)
        self.is_pretty_name_set = False
        self.set_msg(WELCOME_TEXT)

    def set_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.set_info_text(text)

    def append_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.append_info(text)

    def new_task_msg(self, text):
        self.set_msg(text, InfoTypes.TASK)

    def on_change_com_type(self, t):
        dev_type = ComTypes(t)
        if dev_type == ComTypes.SERIAL:
            self.set_msg(
                f'choose {t} if you have found a Baudrate in device documentation!')
        elif dev_type == ComTypes.DLL:
            self.set_msg(
                f'choose {t} if you have file and either a documentation or a .h header file with it')
        elif dev_type == ComTypes.OTHER:
            self.set_msg(
                f"""choose {t} if you your strategy to handle low level communication 
                with a python library from the manufacturer, or from a trusted internet source or have another idea.""")

    def on_pretty_name_set(self):
        self.is_pretty_name_set = True

    def on_create(self):
        self.new_task_msg('creating a new hardware')

        company = self.view.get_company()
        model = self.view.get_model()
        for name, required_field in (('company', company), ('model', model)):
            if required_field == "":
                self.set_msg(f"{name} is required", InfoTypes.FAILED)
                return

        overwrite_existing_module = self.view.get_overwrite()
        dev_type = ComTypes(self.view.get_com_option())
        authors = self.view.get_authors()
        if authors == "":
            authors = 'Mysterion'
            self.view.set_authors(authors)
        pretty_name = self.view.get_pretty_name()
        if pretty_name == "":
            pretty_name = make_pretty_name(model)
            self.view.set_pretty_name(pretty_name)

        for verboten in (" ", "-"):
            if verboten in model:
                self.view.set_model(self._clean_verboten(model, verboten))
                return
            if verboten in company:
                self.view.set_company(self._clean_verboten(company, verboten))
                return
            if verboten in pretty_name:
                self.view.set_pretty_name(
                    self._clean_verboten(pretty_name, verboten))
                return

        try:
            module_name = new_hw(company, model, authors, pretty_name,
                                 overwrite_existing_module, dev_type)
        except FileExistsError as e:
            self.set_msg("Failed: " + str(e), InfoTypes.FAILED)
            return

        self.set_msg(
            f"""created ScopeFoundryHW/{module_name}""",
            InfoTypes.SUCCESS)
        self.append_msg(next_steps(module_name))

    def _clean_verboten(self, text, verboten):
        self.append_msg(
            f'{repr(verboten)} in {text} not allowed: replaced with _', InfoTypes.FAILED)
        return "_".join(text.split(verboten))

    def non_overwrite_guess(self, text=None):
        if not self.is_pretty_name_set:
            self.guess_pretty_name()

    def guess_pretty_name(self):
        model = self.view.get_model()
        self.view.set_pretty_name(make_pretty_name(model))


def make_pretty_name(model: str):
    s = model.split('_')
    if len(s) >= 2:
        name = '_'.join(s[-2:])
    else:
        name = model
    return name.lower()


def next_steps(module_name):
    return f"""<p>typical Next Steps:</p>
            1. run ScopeFoundryHW/{module_name}/*_test_app.py as initial test.<br>
            2. get ScopeFoundryHW/{module_name}/*_dev file to work with your costum read and write functions<br>
            3. in ScopeFoundryHW/{module_name}/*_hw add settings to *_hw file, connect to your read and write functions<br>
            4. test extensivly. <br>
            5. update README.md and publish on GitHub because sharing is caring. <br>
            <br><br>
            Good Luck! <br>
            - UBene.
            """


WELCOME_TEXT = """Although doing some thinking is here is recommended, 
    everything can be changed latter! Defintely have a strategy 
    for your low level communication before you start.
    <br>
    <br>
    a quick google search to see if there already exist a pluggin that you can adapt might 
    save you a lot of time. Use import from github in that case
    """



from ScopeFoundry import BaseApp
import unittest
import time


# see https://bitbucket.org/jmcgeheeiv/pyqttestexample/src/cf039f1a83e390778854f6533adb261d14017d05/src/MargaritaMixerTest.py?at=default&fileviewer=file-view-default

class INI_BaseAppSaveLoadTest(unittest.TestCase):
    
    def setUp(self):
        self.app = BaseApp([])
        
        self.app.settings.New('asdf', dtype=int, initial=12345432334, ro=False, unit='V')
        
        self.ui = self.app.settings.New_UI()
        self.ui.show()
        
        self.ini_fname = "ini_test_%i.ini" % time.time()
        
    def test_save_ini(self):
        self.app.settings_save_ini(self.ini_fname, save_ro=True)
        
    """def test_load_ini(self):
        self.app.settings_load_ini(fname)
    
    """
    def test_round_trip_ini(self):
        self.app.settings_save_ini(self.ini_fname, save_ro=True)
        self.app.settings_load_ini(self.ini_fname)
                        
if __name__ == '__main__':
    unittest.main()
    


from ScopeFoundry import BaseApp
from qtpy import QtWidgets
import pyqtgraph as pg

import logging
logging.basicConfig(level='DEBUG')

class LQRangeTestApp(BaseApp):

    def __init__(self,argv):
        BaseApp.__init__(self,argv)

        ### LQ
        self.x_range = self.settings.New_Range('x', include_center_span=True)#, preserve_num=True)
        self.y_range = self.settings.New_Range('y', include_center_span=True)#, preserve_num=True)

        ### UI
            
        self.ui = QtWidgets.QScrollArea()
        self.ui.setWidgetResizable(True)
        self.ui.setLayout(QtWidgets.QHBoxLayout())

        self.lq_controls = self.settings.New_UI()
        self.ui.layout().addWidget(self.lq_controls)
        
        self.setup_figure()
        
        self.ui.layout().addWidget(self.console_widget)
        #self.plot.show()
        self.ui.show()
        self.ui.setWindowTitle("LQRangeTestApp")
        #self.console_widget.show()
        
        self.ui.setGeometry(300,200,1200,600)
        
    def setup_figure(self):
        self.plot = pg.PlotWidget()
        self.ui.layout().addWidget(self.plot)
        self.plot.showGrid(x=True, y=True)
        self.plot.setAspectLocked(lock=True, ratio=1)

        self.vLine = pg.InfiniteLine(angle=90, movable=False)
        self.hLine = pg.InfiniteLine(angle=0, movable=False)
        self.plot.addItem(self.vLine, ignoreBounds=True)
        self.plot.addItem(self.hLine, ignoreBounds=True)


        self.scan_roi = pg.ROI([0,0],[1,1], movable=True)
        self.scan_roi.addScaleHandle([1, 1], [0, 0])
        self.scan_roi.addScaleHandle([0, 0], [1, 1])
        self.update_scan_roi()
        self.scan_roi.sigRegionChangeFinished.connect(self.mouse_update_scan_roi)
        self.plot.addItem(self.scan_roi)        

        #
        for lq in self.x_range.lq_dict.values():       
            lq.updated_value.connect(self.update_scan_roi)
        for lq in self.y_range.lq_dict.values():       
            lq.updated_value.connect(self.update_scan_roi)



    def update_scan_roi(self):
        self.log.debug("update_scan_roi")
        X = self.settings.ranges['x']
        Y = self.settings.ranges['y']
        self.imshow_extent = [X.min.val - 0.5*X.step.val,
                              X.max.val + 0.5*X.step.val,
                              Y.min.val - 0.5*Y.step.val,
                              Y.max.val + 0.5*Y.step.val]
        x0, x1, y0, y1 = self.imshow_extent
        
        self.vLine.setPos(X.center.val)
        self.hLine.setPos(Y.center.val)
        
        self.scan_roi.blockSignals(True)
        self.scan_roi.setPos( (x0, y0, 0))
        self.scan_roi.setSize( (x1-x0, y1-y0, 0))
        self.scan_roi.blockSignals(False)

    def mouse_update_scan_roi(self):
        x0,y0 =  self.scan_roi.pos()
        w, h =  self.scan_roi.size()
        
        self.x_range.center.update_value(x0 + w/2)
        self.y_range.center.update_value(y0 + h/2)
        self.x_range.span.update_value(w-self.x_range.step.val)
        self.y_range.span.update_value(h-self.y_range.step.val)
        #self.compute_scan_params()
        self.update_scan_roi()

        
if __name__ == '__main__':
    app = LQRangeTestApp([])
    app.exec_()


'''
Created on Sep 17, 2021

@author: Benedikt Ursprung
'''
import logging
import sys

from ScopeFoundry.base_app import BaseMicroscopeApp

level = logging.DEBUG
logging.basicConfig(level=level)


class App(BaseMicroscopeApp):

    name = 'sequencer_test_app'

    def setup(self):

        from ScopeFoundry.sequencer import Sequencer, SweepSequencer
        self.add_measurement(Sequencer(self))
        self.add_measurement(SweepSequencer(self))

if __name__ == '__main__':
    app = App(sys.argv)
    sys.exit(app.exec_())



from __future__ import absolute_import, print_function
import functools
import time
from datetime import datetime
import os

import h5py

from .cb32_uuid import cb32_uuid

"""
recommended HDF5 file format for ScopeFoundry
* = group
- = attr
D = data_set

* /
    - scope_foundry_version = 130
    - time_id = # unix epoch timestamp
    - unique_id = # persistent identifer for dataset, 26 character
    - uuid = # uuid string representation of unique_id
    * app
        - ScopeFoundry_Type = App
        - name = test_app
        * settings
            - log_quant_1
            - log_quant_1_unit
            - ...
    * hardware
        * hardware_component_1
            - ScopeFoundry_Type = Hardware
            - name = hardware_component_1
            * settings
                - log_quant_1
                - ...
                * units
                    - log_quant_1 = '[n_m]'
        * ...
    * measurement
        * measurement_1
            - ScopeFoundry_Type = Measurement
            - name = measurement_1
                * settings
                    - log_quant_1
                    - ...
                    * units
                        - log_quant_1 = '[n_m]'
            D simple_data_set_2
            D ...

other thoughts:
    store git revision of code
    store git revision of ScopeFoundry
    EMD compatibility, NeXUS compatibility
"""


def h5_base_file(app, fname=None, measurement=None):
    unique_id, u = cb32_uuid() # persistent identifier for dataset
    t0 = time.time()
    if fname is None and measurement is not None:
        
        f = app.settings['data_fname_format'].format(
            app=app,
            measurement=measurement,
            timestamp=datetime.fromtimestamp(t0),
            unique_id=unique_id,
            unique_id_short = unique_id[0:13],
            ext='h5')
        fname = os.path.join(app.settings['save_dir'], f)        
        #fname = os.path.join(app.settings['save_dir'], "%i_%s.h5" % (t0, measurement.name) )
    h5_file = h5py.File(fname, 'a')
    root = h5_file['/']
    root.attrs["ScopeFoundry_version"] = 130
    root.attrs['time_id'] = int(t0)
    root.attrs['unique_id'] = unique_id
    root.attrs['uuid'] = str(u)

    h5_save_app_lq(app, root)
    h5_save_hardware_lq(app, root)
    return h5_file

def h5_save_app_lq(app, h5group):
    h5_app_group = h5group.create_group('app/')
    h5_app_group.attrs['name'] = app.name
    h5_app_group.attrs['ScopeFoundry_type'] = "App"
    settings_group = h5_app_group.create_group('settings')
    h5_save_lqcoll_to_attrs(app.settings, settings_group)

def h5_save_hardware_lq(app, h5group):
    h5_hardware_group = h5group.create_group('hardware/')
    h5_hardware_group.attrs['ScopeFoundry_type'] = "HardwareList"
    for hc_name, hc in app.hardware.items():
        h5_hc_group = h5_hardware_group.create_group(hc_name)
        h5_hc_group.attrs['name'] = hc.name
        h5_hc_group.attrs['ScopeFoundry_type'] = "Hardware"
        h5_hc_settings_group = h5_hc_group.create_group("settings")
        h5_save_lqcoll_to_attrs(hc.settings, h5_hc_settings_group)
    return h5_hardware_group

def h5_save_lqcoll_to_attrs(settings, h5group):
    """
    take a LQCollection
    and create attributes inside h5group

    :param logged_quantities:
    :param h5group:
    :return: None
    """
    unit_group = h5group.create_group('units')
    # TODO decide if we should specify h5 attr data type based on LQ dtype
    for lqname, lq in settings.as_dict().items():
        #print('h5_save_lqcoll_to_attrs', lqname, repr(lq.val))
        try:
            h5group.attrs[lqname] = lq.val
        except:
            h5group.attrs[lqname] = lq.ini_string_value()
        if lq.unit:
            unit_group.attrs[lqname] = lq.unit


def h5_create_measurement_group(measurement, h5group, group_name=None):
    if group_name is None:
        group_name = 'measurement/' + measurement.name
    h5_meas_group = h5group.create_group(group_name)
    h5_save_measurement_settings(measurement, h5_meas_group)
    return h5_meas_group

def h5_save_measurement_settings(measurement, h5_meas_group):
    h5_meas_group.attrs['name'] = measurement.name
    h5_meas_group.attrs['ScopeFoundry_type'] = "Measurement"
    settings_group = h5_meas_group.create_group("settings")
    h5_save_lqcoll_to_attrs(measurement.settings, settings_group)
    
    
def h5_measurement_file(measurement,  fname=None):
    """ Default way to create HDF5 file and fill with 
    metadata for measurement and hardware
    
    filename of file is determined by app settings unless fname is specified
    
    creates an h5 file with groups: /app, /hardware, /measurement/<measurement_name>
    
    with all settings written to the file
    
    returns Measurement H5 group where additional data objects can be added during the measurment
    """
    h5f = h5_base_file(
                    app=measurement.app,
                    fname=fname,
                    measurement=measurement)
    M = h5_create_measurement_group(measurement, h5group=h5f)
    return M
    
    
def h5_create_emd_dataset(name, h5parent, shape=None, data = None, maxshape = None, 
                          dim_arrays = None, dim_names= None, dim_units = None,  **kwargs):
    """
    create an EMD dataset v0.2 inside h5parent
    returns an h5 group emd_grp
    
    to access N-dim dataset:    emd_grp['data']
    to access a specific dimension array: emd_grp['dim1']

    HDF5 Hierarchy:
    ---------------
    * h5parent
        * name [emd_grp] (<--returned)
            - emd_group_type = 1
            D data [shape = shape] 
            D dim1 [shape = shape[0]]
                - name
                - units
            ...
            D dimN [shape = shape[-1]]      

    Parameters
    ----------
    
    h5parent : parent HDF5 group 
    
    shape : Dataset shape of N dimensions.  Required if "data" isn't provided.

    data : Provide data to initialize the dataset.  If used, you can omit
            shape and dtype arguments.
    
    Keyword Args:
    
    dtype : Numpy dtype or string.  If omitted, dtype('f') will be used.
            Required if "data" isn't provided; otherwise, overrides data
            array's dtype.
            
    dim_arrays : optional, a list of N dimension arrays
    
    dim_names : optional, a list of N strings naming the dataset dimensions 
    
    dim_units : optional, a list of N strings specifying units of dataset dimensions
    
    Other keyword arguments follow from h5py.File.create_dataset
    
    Returns
    -------
    emd_grp : h5 group containing dataset and dimension arrays, see hierarchy below
    
    """
    #set the emd version tag at root of h5 file
    h5parent.file['/'].attrs['version_major'] = 0
    h5parent.file['/'].attrs['version_minor'] = 2
        
    # create the EMD data group
    emd_grp = h5parent.create_group(name)
    emd_grp.attrs['emd_group_type'] = 1
    
    if data is not None:
        shape = data.shape
    
    # data set where the N-dim data is stored
    data_dset = emd_grp.create_dataset("data", shape=shape, maxshape=maxshape, data=data, **kwargs)
    
    if dim_arrays is not None: assert len(dim_arrays) == len(shape)
    if dim_names  is not None: assert len(dim_names)  == len(shape)
    if dim_units  is not None: assert len(dim_units)  == len(shape)
    if maxshape   is not None: assert len(maxshape)   == len(shape)
    
    # Create the dimension array datasets
    for ii in range(len(shape)):
        if dim_arrays is not None:
            dim_array = dim_arrays[ii]
            dim_dtype =  dim_array.dtype            
        else:
            dim_array = None
            dim_dtype = float
        if dim_names is not None:
            dim_name = dim_names[ii]
        else:
            dim_name = "dim" + str(ii+1)
        if dim_units is not None:
            dim_unit = dim_units[ii]
        else:
            dim_unit = None
        if maxshape is not None:
            dim_maxshape = (maxshape[ii],)
        else:
            dim_maxshape = None
        
        # create dimension array dataset
        dim_dset = emd_grp.create_dataset("dim" + str(ii+1), shape=(shape[ii],), 
                                           dtype=dim_dtype, data=dim_array, 
                                           maxshape=dim_maxshape)
        dim_dset.attrs['name'] = dim_name
        if dim_unit is not None:
            dim_dset.attrs['unit'] = dim_unit
            
    return emd_grp
    

def create_extendable_h5_dataset(h5_group, name, shape, axis=0, dtype=None, **kwargs):
    """
    Create and return an empty HDF5 dataset of type *dtype* in h5_group that can store
    an infinitely long log of along *axis* (defaults to axis=0). 
    Dataset will have an initial shape *shape* but can be extended along *axis*
            
    creates reasonable defaults for chunksize
    can be overridden with **kwargs that are sent directly to 
    h5_group.create_dataset
    """
    maxshape = list(shape)
    maxshape[axis] = None
    
    default_kwargs = dict(
        name=name,
        shape=shape,
        dtype=dtype,
        #chunks=(1,),
        chunks=shape,
        maxshape=maxshape,
        compression=None,
        #shuffle=True,
        )
    default_kwargs.update(kwargs)
    h5_dataset =  h5_group.create_dataset(
        **default_kwargs
        )
    return h5_dataset
    
    
def create_extendable_h5_like(h5_group, name, arr, axis=0, **kwargs):
    """
    Create and return an empty HDF5 dataset in h5_group that can store
    an infinitely long log of along *axis* (defaults to axis=0). Dataset will be the same
    shape as *arr* but can be extended along *axis*
            
    creates reasonable defaults for chunksize, and dtype,
    can be overridden with **kwargs that are sent directly to 
    h5_group.create_dataset
    """
    return create_extendable_h5_dataset(
        h5_group, name, arr.shape, axis, arr.dtype, **kwargs)

def extend_h5_dataset_along_axis(ds, new_len, axis=0):
    newshape = list(ds.shape)
    newshape[axis] = new_len
    ds.resize( newshape )    

    
def load_settings(fname):
    """
    returns a dictionary (path, value) of all settings stored in a h5 file
    """
    if not fname.endswith(".h5"):
        return {}
    
    settings = {}
    visit_func = functools.partial(_settings_visitfunc, settings=settings)
    
    with h5py.File(fname) as file:
        file.visititems(visit_func)    

    return settings


def _settings_visitfunc(name, node, settings):
    if not name.endswith("settings"):
        return
    
    for key, val in node.attrs.items():
        lq_path = f"{name.replace('settings', key)}"
        settings[lq_path] = val



from setuptools import setup

setup(
    name = 'ScopeFoundry',
    
    version = '1.5.0',
    
    description = 'a platform for laboratory equipment control and scientific data analysis',
    long_description =open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    
    # Author details
    author='Edward S. Barnard',
    author_email='esbarnard@lbl.gov',

    # Choose your license
    license='BSD',
    
    url='http://www.scopefoundry.org/',

    package_dir={'ScopeFoundry': '.'},
    
    packages=['ScopeFoundry',
              'ScopeFoundry.controlling',
              'ScopeFoundry.data_browser',
                  'ScopeFoundry.data_browser.viewers',
                  'ScopeFoundry.data_browser.plug_ins',
              'ScopeFoundry.examples',
              'ScopeFoundry.graphics',
              'ScopeFoundry.graphics.zoomable_map',
              'ScopeFoundry.plugin_manager',  
              'ScopeFoundry.plugin_manager.features',
              'ScopeFoundry.plugin_manager.templates',
              'ScopeFoundry.scanning',
              'ScopeFoundry.sequencer',
              'ScopeFoundry.sequencer.item_types',              
            ],
    
    #packages=find_packages('.', exclude=['contrib', 'docs', 'tests']),
    #include_package_data=True,  
    
    package_data={
        '':["*.ui", "*.icns", '*.png', '*.svg'], # include QT ui files and logo and icons
        },
        
    install_requires = [
    	'numpy', 'h5py', 'qtpy', 'xreload', 'uuid7'],
    
    extras_require={
        'all' : ['qtconsole', 'pyqtdarktheme'],
    }

    )



from pathlib import Path

from .publish_on_gh_view import PublishOnGhView as View
from .utils import InfoTypes, decorate_info_text, run_subprocess


class PublishOnGhController:
    def __init__(self, view: View):
        self.view = view
        view.update_hw_options_clicked(self.update_hw_options)
        view.subtree_clicked(self.subtree_hw)
        view.publish_clicked(self.publish_on_gh)
        self.update_hw_options()

    def set_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.set_info_text(text)

    def append_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.append_info(text)

    def new_task_msg(self, text):
        self.set_msg(text, InfoTypes.TASK)

    def update_hw_options(self):
        hw_path = Path('../../ScopeFoundryHW/')
        if not hw_path.is_dir():
            hw_path.mkdir()
        self.view.set_hw_options(
            [p.name for p in hw_path.iterdir()])

    def subtree_hw(self):
        name = self.view.get_hw()
        module_name = f"HW_{name}"

        self.new_task_msg(f'attempting to subtree {module_name}')
        path = Path(f'../my_hw_repos/{module_name}.git')
        if path.is_dir():
            self.append_msg(
                f'Path {path} already exists use push instead', InfoTypes.FAILED)
            return

        stdout, stderr = subtree_hw(name)
        self.append_msg(stdout, InfoTypes.INFO)
        self.append_msg(stderr, InfoTypes.ERROR)
        # self.publish_on_gh()

    def publish_on_gh(self):

        name = self.view.get_hw()
        module_name = f"HW_{name}"
        my_gh_acc = self.view.get_my_gh_acc()
        gh_repo = f"https://github.com/{my_gh_acc}/{module_name}.git"

        if self.repo_exist(gh_repo):
            self.append_msg(
                f"repo already exists: {gh_repo}", InfoTypes.FAILED)
            return

        self.new_task_msg(f'publish {gh_repo}')
        stdout, stderr = publish_on_gh(module_name, gh_repo)
        self.append_msg(stdout, InfoTypes.INFO)
        self.append_msg(stderr, InfoTypes.ERROR)

    def repo_exist(self, repo):
        stdout, stderr = ls_remote(repo)
        return stderr == ""


def ls_remote(repo):
    cmd = f"git ls-remote -h {repo}"
    return run_subprocess(cmd)


def subtree_hw(name):
    cmd = f"bash scripts/subtree_hw.sh {name}"
    return run_subprocess(cmd)


def publish_on_gh(module_name, gh_repo):
    cmd = f"bash scripts/publish_on_gh.sh {module_name} {gh_repo}"
    return run_subprocess(cmd)



from qtpy.QtWidgets import QHBoxLayout

from .sequencer import Sequencer


class SweepSequencer(Sequencer):
    name = "sweep_sequencer"

    def setup(self):
        Sequencer.setup(self)
        self.range = self.settings.New_Range(
            "range",
            include_sweep_type=True,
            initials=[67, 410, 10],
            description=RANGEDESCRIPTION,
            spinbox_decimals=5,
        )
        self.settings.New("current_range_value", ro=True)
        self.settings.New("ignore_sweep", bool, initial=False)

    def setup_figure(self):
        Sequencer.setup_figure(self)
        layout = QHBoxLayout()
        layout.addWidget(self.range.New_UI())
        layout.addWidget(self.settings.New_UI(["current_range_value", "ignore_sweep"]))
        self.layout.insertLayout(1, layout)
        try:
            self.editors["update-setting"].ui.value_le.setText(
                "measurement/sweep_sequencer/current_range_value"
            )
        except AttributeError:
            pass

    def run(self):
        if self.settings["ignore_sweep"]:
            Sequencer.run(self)
        else:
            values = self.range.sweep_array
            for i, x in enumerate(values):
                if self.interrupt_measurement_called:
                    break
                self.settings["current_range_value"] = x
                # print(self.name, 'current_range_value', x)
                self.set_progress(100.0 * i / len(values))
                Sequencer.run(self)


RANGEDESCRIPTION = """use <i>measurement/sweep_sequencer/current_range_value</i> 
to update the setting you want to sweep"""



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''
from ScopeFoundry.hardware import HardwareComponent


class $HW_CLASS_NAME(HardwareComponent):

    name = "$HW_NAME"

    def setup(self):
        S = self.settings
        S.New("port", str, initial="COM1",
              description='COMx see device manager')
        S.New("property_x", float, ro=False, unit='mm', spinbox_decimals=4, si=False)


    def connect(self):

        S = self.settings

        $IMPORT_DEV
        self.dev = $DEV_CLASS_NAME(S["port"], debug=S['debug_mode'])

        # attache read and write functions to settings
        S.property_x.connect_to_hardware(
            self.dev.read_property_x,
            self.dev.write_property_x)

    def disconnect(self):
        if not hasattr(self, 'dev'):
            return

        # self.settings.disconnect_all_from_hardware()
        self.dev.close()
        del self.dev

    # if you want to continuously update settings implement *run* method
    # def run(self):
    #     self.settings.property_x.read_from_hardware()
